import logging
from copy import deepcopy
from math import ceil
from typing import Union, Dict, List

import numpy as np
from datasets.arrow_dataset import Dataset as ArrowDataset
from sklearn.model_selection import train_test_split
from torch.utils.data import Dataset
from transformers import PreTrainedTokenizerFast
from transformers.tokenization_utils_base import BatchEncoding

log = logging.getLogger()


class TransformersDataset(Dataset):
    def __init__(
        self,
        instances: Union[list, ArrowDataset, Dict[str, List[Union[list, str, int]]]],
        text_column_name: str = "text",
        label_column_name: str = "label",
        tokenizer: PreTrainedTokenizerFast or None = None,
        tokenization_kwargs: dict or None = None,
        task: str = "cls",  # or "ner"
        id2label: Dict[int, str] = None,
        label_smoothing: bool = False,  # whether labels are smoothed
    ):
        """
        Class, immitating ArrowDataset from HuggingFace datasets
        :param instances: values of the dataset
        :param text_column_name:
        :param label_column_name:
        :param tokenizer:
        :param tokenization_kwargs:
        :param task:
        :param id2label:
        :param label_smoothing:
        """
        super().__init__()
        if isinstance(instances, (dict, BatchEncoding)):
            self._init_from_dict_with_instances(instances)
        else:
            self._init_from_list_of_instances_or_arrow_dataset(instances)

        self.text_column_name = text_column_name
        self.label_column_name = label_column_name
        # For compatibility with ArrowDataset
        self.features = {k: None for k in self.instances[0]}
        self.num_rows = len(self)

        self.tokenizer = tokenizer
        self.tokenization_kwargs = (
            tokenization_kwargs if tokenization_kwargs is not None else {}
        )
        if task == "ner":
            self.tokenization_kwargs["is_split_into_words"] = True
        self.tokenized_data = None

        self.task = task
        self.id2label = id2label
        self.label_smoothing = label_smoothing

    def __len__(self):
        return len(self.instances)

    def __getitem__(
        self, key: Union[int, slice, list, str]
    ) -> Dict[Union[str, int], list]:

        if isinstance(key, str):
            return self.columns_dict[key].tolist()
        elif isinstance(key, int):
            return self._getitem_int(key)
        return self._getitem_list_or_slice(key)

    def select(self, key: Union[int, slice, list]) -> "TransformersDataset":
        if isinstance(key, int):
            instances = [self[key]]
        else:
            instances = self[key]

        return TransformersDataset(
            instances,
            text_column_name=self.text_column_name,
            label_column_name=self.label_column_name,
            tokenizer=self.tokenizer,
            tokenization_kwargs=self.tokenization_kwargs,
            task=self.task,
            id2label=self.id2label,
        )

    def add(
        self,
        instance_or_instances: Union[list, dict, "TransformersDataset"],
        inplace=True,
    ) -> Union[None, "TransformersDataset"]:

        if isinstance(instance_or_instances, dict):
            instance_or_instances = [instance_or_instances]
        elif isinstance(instance_or_instances, TransformersDataset):
            instance_or_instances = instance_or_instances.instances

        columns = self.columns
        # Need to sort columns the same way they are done in `instances` to prevent cases [0, 1] != [1, 0]
        if sorted(columns) == sorted(instance_or_instances[0].keys()):
            columns = list(instance_or_instances[0].keys())
        else:
            raise RuntimeError(
                "Instance columns do not match with the Dataset columns!"
            )
        if inplace:
            self.instances = self.instances + list(instance_or_instances)
            # First convert back to list and then will convert back to np.ndarray to accelerate the process
            for column in columns:
                self.columns_dict[column] = list(self.columns_dict[column])
            for inst in instance_or_instances:
                assert (
                    list(inst.keys()) == columns
                ), "Instance columns do not match with the Dataset columns!"
                for column in columns:
                    self.columns_dict[column].append(inst[column])
            # Convert back to np.ndarray
            for column in columns:
                self.columns_dict[column] = np.array(self.columns_dict[column])
        else:
            dataset = deepcopy(self)
            dataset.instances = dataset.instances + list(instance_or_instances)
            # First convert back to list and then will convert back to np.ndarray to accelerate the process
            for column in columns:
                dataset.columns_dict[column] = list(dataset.columns_dict[column])
            for inst in instance_or_instances:
                assert (
                    list(inst.keys()) == columns
                ), "Instance columns do not match with the Dataset columns!"
                for column in columns:
                    dataset.columns_dict[column].append(inst[column])
            # Convert back to np.ndarray
            for column in columns:
                dataset.columns_dict[column] = np.array(dataset.columns_dict[column])

            return dataset

    def add_column(self, column_name: str, column_data: Union[list, np.ndarray]):
        for i, new_column_data in enumerate(column_data):
            self.instances[i][column_name] = new_column_data
        self.columns.append(column_name)
        if isinstance(column_data, list):
            self.columns_dict[column_name] = np.array(column_data)
        else:
            self.columns_dict[column_name] = column_data

    def _getitem_int(self, idx: int) -> Dict[Union[str, int], list]:
        instance = self.instances[idx]
        if self.tokenizer is None:
            return instance

        encoded = self.tokenizer(
            instance[self.text_column_name], **self.tokenization_kwargs
        )
        encoded.update(instance)
        return encoded

    def _getitem_list_or_slice(
        self, list_or_slice_idx: Union[List[int], slice]
    ) -> Dict[Union[str, int], list]:

        instances_dict = self._get_return_dict_for_list_or_slice(list_or_slice_idx)

        if self.tokenizer is None:
            return instances_dict

        # If there is a tokenizer, tokenize the text data and add it to the return dict
        texts = self.columns_dict[self.text_column_name]
        texts_idx = texts[list_or_slice_idx].tolist()
        encoded = self.tokenizer(texts_idx, **self.tokenization_kwargs)
        encoded.update(instances_dict)
        return encoded

    def _get_return_dict_for_list_or_slice(
        self, list_or_slice_idx: Union[List[int], slice]
    ) -> Dict[Union[str, int], list]:
        return {
            column: self.columns_dict[column][list_or_slice_idx].tolist()
            for column in self.columns
        }

    def shuffle(self, seed=42):
        np.random.seed(seed)
        idx = np.arange(len(self.instances))
        np.random.shuffle(idx)
        self.instances = [self.instances[x] for x in idx]
        for key, value in self.columns_dict.items():
            self.columns_dict[key] = value[idx]
        return self

    # def tokenize_data(
    #     self,
    #     tokenizer: PreTrainedTokenizerFast or None = None,
    #     text_column_name: str or None = None,
    #     label_column_name: str or None = None,
    #     save_first_bpe_mask: bool = False,
    #     ignore_tokenized_data: bool = True,
    #     **kwargs,
    # ):
    #     if self.tokenized_data is not None and not ignore_tokenized_data:
    #         return self.tokenized_data
    #
    #     tokenizer = tokenizer if tokenizer is not None else self.tokenizer
    #     tokenization_kwargs = deepcopy(self.tokenization_kwargs)
    #     tokenization_kwargs.update(kwargs)
    #     # Add truncation if it were not specified that is must be disabled
    #     if "truncation" not in tokenization_kwargs:
    #         tokenization_kwargs["truncation"] = True
    #
    #     text_column_name = (
    #         text_column_name if text_column_name is not None else self.text_column_name
    #     )
    #     label_column_name = (
    #         label_column_name
    #         if label_column_name is not None
    #         else self.label_column_name
    #     )
    #
    #     texts = list(self.columns_dict[text_column_name])
    #     labels = list(self.columns_dict[label_column_name])
    #     # TODO: comb this part
    #     if len(self) < 150_000 and not self.task.startswith("cv"):
    #         if self.task == "cls":
    #             batch_encoding = tokenizer(texts, **tokenization_kwargs)
    #         elif self.task == "ner":
    #             for redundant_key in [
    #                 "is_split_into_words",
    #                 "return_offsets_mapping",
    #             ]:
    #                 if redundant_key in tokenization_kwargs:
    #                     del tokenization_kwargs[redundant_key]
    #
    #             batch_encoding = tokenizer(
    #                 texts,
    #                 is_split_into_words=True,
    #                 **tokenization_kwargs,
    #             )
    #             tags, data_idx_first_bpe = [], []
    #             for i in range(len(batch_encoding.encodings)):
    #                 encoding_labels = align_labels_with_tokens(
    #                     labels[i], batch_encoding.word_ids(i)
    #                 )
    #                 tags.append(encoding_labels)
    #                 # Save ids of the first BPEs if necessary
    #                 if save_first_bpe_mask:
    #                     word_ids = batch_encoding.word_ids(i)
    #                     idx_first_bpe = [
    #                         i
    #                         for (i, x) in enumerate(word_ids[1:], 1)
    #                         if ((x != word_ids[i - 1]) and (x is not None))
    #                     ]
    #                     data_idx_first_bpe.append(idx_first_bpe)
    #             # Dummy step to make the next step equal for both tasks
    #             labels = tags
    #         else:
    #             raise NotImplementedError
    #         # Since we get batch encoding instead of list of tokenized instances, need to reformat it
    #         tokenized_data = [
    #             {
    #                 "input_ids": inst.ids,
    #                 "attention_mask": inst.attention_mask,
    #                 "labels": labels[i],
    #             }
    #             for i, inst in enumerate(batch_encoding.encodings)
    #         ]
    #         if save_first_bpe_mask and self.task == "ner":
    #             [
    #                 tokenized_data[i].update({"idx_first_bpe": data_idx_first_bpe[i]})
    #                 for i in range(len(tokenized_data))
    #             ]
    #
    #         self.tokenized_data = tokenized_data
    #     elif self.task.startswith("cv"):
    #         self.tokenized_data = [
    #             {"features": text, "labels": label}
    #             for text, label in zip(texts, labels)
    #         ]
    #     else:
    #         raise NotImplementedError("Not implemented for more than 150k instances")
    #
    #     return TransformersDataset(
    #         tokenized_data,
    #         text_column_name=self.text_column_name,
    #         label_column_name=self.label_column_name,
    #         tokenizer=self.tokenizer,
    #         tokenization_kwargs=self.tokenization_kwargs,
    #         task=self.task,
    #         id2label=self.id2label,
    #     )

    def __repr__(self):
        return f"TransformersDataset with num rows = {len(self)} and columns = {self.columns}"

    def train_test_split(self, train_size=None, test_size=None, shuffle=False, seed=42):

        if train_size is None and test_size is None:
            test_size = 0.2
        elif train_size is not None and test_size is None:
            test_size = 1 - train_size
        elif train_size is not None and test_size is not None:
            log.warning(
                "Both `train_size` and `test_size` are provided. Ignoring `train_size`."
            )

        ids = np.arange(len(self))
        train_ids, test_ids = train_test_split(
            ids, test_size=test_size, shuffle=shuffle, random_state=seed
        )

        train_data = self.select(train_ids)
        test_data = self.select(test_ids)

        return {"train": train_data, "test": test_data}

    def _init_from_list_of_instances_or_arrow_dataset(
        self, instances: Union[list, ArrowDataset],
    ):
        columns = instances[0].keys()
        self.columns = list(columns)
        # Check all instances have the same columns
        assert all(
            [inst.keys() == columns for inst in instances]
        ), "All the instances must have the same keys!"

        self.instances = list(instances)
        self.columns_dict = {}
        for column in self.columns:
            self.columns_dict[column] = np.array([inst[column] for inst in instances])

    def _init_from_dict_with_instances(
        self, instances: Dict[str, List[Union[list, str, int]]],
    ):
        columns = instances.keys()
        self.columns = list(columns)
        # Check all columns have the same length
        first_column_length = len(instances[self.columns[0]])
        assert all(
            [
                len(instances[column]) == first_column_length
                for column in self.columns[1:]
            ]
        ), "All columns must have the same length!"

        self.instances = [
            {column: instances[column][i] for column in self.columns}
            for i in range(len(instances[self.columns[0]]))
        ]
        self.columns_dict = {}
        for column in columns:
            self.columns_dict[column] = np.array(instances[column])

    def __delitem__(self, key):
        if isinstance(key, str):
            del self.columns_dict[key]
            self.columns.remove(key)
            [inst.pop(key) for inst in self.instances]
        elif isinstance(key, int):
            self.instances.pop(key)
            for col in self.columns_dict:
                self.columns_dict[col] = np.delete(self.columns_dict[col], key)
        elif isinstance(key, (list, tuple, np.ndarray)) and isinstance(key[0], int):
            for i in key:
                self.__delitem__(i)
        else:
            raise NotImplementedError

    def remove_columns_(self, columns):
        if isinstance(columns, str):
            self.__delitem__(columns)
        else:
            for column in columns:
                self.__delitem__(column)

    def remove_columns(self, columns):
        if isinstance(columns, str):
            columns = [columns]

        data_without_req_columns = []
        for x in self.instances:
            inst = {k: v for k, v in x.items() if k not in columns}
            data_without_req_columns.append(inst)
        return TransformersDataset(
            data_without_req_columns,
            task=self.task,
            text_column_name=self.text_column_name,
            label_column_name=self.label_column_name,
        )

    def rename_column(self, original_column_name: str, new_column_name: str):
        dataset = deepcopy(self)
        for i, inst in enumerate(dataset.instances):
            dataset.instances[i][new_column_name] = dataset.instances[i][
                original_column_name
            ]
            dataset.instances[i].pop(original_column_name)

        dataset.columns_dict[new_column_name] = dataset.columns_dict[
            original_column_name
        ]
        dataset.columns_dict.pop(original_column_name)
        return dataset

    def map(
        self,
        function,
        batched: bool = False,
        remove_columns: Union[List[str], str, None] = None,
        batch_size: int = 1000,
        **kwargs,
    ):
        instances_to_tokenize = []
        for x in self.instances:
            inst = {}
            for key, value in x.items():
                if isinstance(value, np.ndarray):
                    value = list(value)
                inst[key] = value
            instances_to_tokenize.append(inst)

        num_batches = ceil(len(instances_to_tokenize) / batch_size)
        # TODO: parallelize
        instances_tokenized = []
        for i_batch in range(num_batches):
            instances_batch = self._concatenate_instances(
                instances_to_tokenize[i_batch * batch_size : (i_batch + 1) * batch_size]
            )
            output = function(instances_batch)
            if isinstance(output, BatchEncoding):
                instances_tokenized.append(output.data)
            else:
                raise NotImplementedError

        instances = self._concatenate_batch_encodings_data(instances_tokenized)

        dataset = TransformersDataset(
            instances,
            text_column_name=self.text_column_name,
            label_column_name=self.label_column_name,
            tokenization_kwargs=self.tokenization_kwargs,
            task=self.task,
            id2label=self.id2label,
        )
        if remove_columns is not None:
            columns_to_remove = []
            for x in remove_columns:
                if x in dataset.columns:
                    columns_to_remove.append(x)
            dataset.remove_columns_(columns_to_remove)

        return dataset

    @staticmethod
    def _concatenate_instances(instances: List[dict]):
        keys = instances[0].keys()
        return {key: [inst[key] for inst in instances] for key in keys}

    @staticmethod
    def _concatenate_batch_encodings_data(data: List[Dict]):
        keys = data[0].keys()
        return {key: [obj for inst in data for obj in inst[key]] for key in keys}



import numpy as np


def naive_subsampling(uncertainty_estimates, gamma_or_k_confident_to_save, **kwargs):
    if isinstance(gamma_or_k_confident_to_save, float):
        gamma_or_k_confident_to_save = int(
            gamma_or_k_confident_to_save * len(uncertainty_estimates)
        )
    argsort = np.argsort(-uncertainty_estimates)
    return argsort[:gamma_or_k_confident_to_save]



import gc
import time
from tqdm import tqdm
from typing import Dict
import numpy as np
from abc import abstractmethod
from pathlib import Path
from requests.models import HTTPError

import torch
from torch.utils.data import DataLoader
from transformers import (
    DataCollatorForSeq2Seq,
    Seq2SeqTrainingArguments,
    Seq2SeqTrainer,
    set_seed,
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
)

from .transformers_base_wrapper import TransformersBaseWrapper
from ...utils.general import tensor_to_numpy

import logging

log = logging.getLogger()


class WrapperSeq2Seq(TransformersBaseWrapper):
    @staticmethod
    def get_model_class():
        return AutoModelForSeq2SeqLM

    @staticmethod
    def get_training_args_class():
        return Seq2SeqTrainingArguments

    @staticmethod
    def get_data_collator_class():
        return DataCollatorForSeq2Seq

    @staticmethod
    def get_trainer_class():
        return Seq2SeqTrainer

    def get_additional_training_kwargs(self, train_data):
        generation_max_length = (
            self._get_generation_max_length(train_data, self.data_config)
            if self._trainer_kwargs.generation_max_length is None
            else self._trainer_kwargs.generation_max_length
        )
        log.info(f"Using generation max length == {generation_max_length}")
        generation_num_beams = self._trainer_kwargs.generation_num_beams
        return {
            "predict_with_generate": True,
            "include_inputs_for_metrics": True,
            "generation_max_length": generation_max_length,
            "generation_num_beams": generation_num_beams,
        }

    def get_model_and_tokenizer(
        self, model_cfg, seed: int = 42, cache_dir=None, **kwargs
    ):
        set_seed(seed)
        pretrained_model_name = model_cfg.checkpoint

        model_cache_dir = Path(cache_dir) / "model" if cache_dir is not None else None
        tokenizer_cache_dir = (
            Path(cache_dir) / "tokenizer" if cache_dir is not None else None
        )
        ### Model part
        model_class = self.get_model_class()
        try:
            model = model_class.from_pretrained(
                pretrained_model_name, cache_dir=model_cache_dir,
            )
        # If there are troubles with the connection, try to build from local files
        except HTTPError:
            log.warning(
                f"Connection troubles. Using local model {pretrained_model_name}."
            )
            model = model_class.from_pretrained(
                pretrained_model_name, cache_dir=model_cache_dir, local_files_only=True,
            )
        ### Tokenizer part
        try:
            tokenizer = AutoTokenizer.from_pretrained(
                pretrained_model_name, cache_dir=tokenizer_cache_dir
            )
        except HTTPError:
            tokenizer = AutoTokenizer.from_pretrained(
                pretrained_model_name,
                cache_dir=tokenizer_cache_dir,
                local_files_only=True,
            )
        if model_cfg.tokenizer_max_length is not None:
            tokenizer.model_max_length = model_cfg.tokenizer_max_length
        return model, tokenizer

    def get_predictions(
        self,
        data,
        is_tokenized=False,
        data_config=None,
        calculate_time=False,
        evaluate: bool = False,
        use_predict_loop: bool = False,
    ):
        if data_config is None:
            data_config = self.data_config
        text_name = data_config["text_name"]
        label_name = data_config["label_name"]

        if not is_tokenized:
            data = self.tokenize_data(
                tokenizer=self.tokenizer,
                data=data,
                text_name=text_name,
                label_name=label_name,
                test_mode=not evaluate,
            )
        assert (
            getattr(self, "trainer", None) is not None
        ), "Instantiate the Trainer first!"
        start_time = time.time()
        predictions = self.trainer.predict(data)

        if calculate_time:
            self._calculate_time(start_time, phase="predict")
        torch.cuda.empty_cache()
        gc.collect()

        return predictions

    def generate(
        self,
        data,
        return_scores: bool = True,
        return_decoded_preds: bool = True,
        is_tokenized: bool = False,
        data_config=None,
        to_numpy: bool = False,
        to_eval_mode: bool = True,
        batch_size: int = None,
        calculate_time: bool = True,
        generation_max_length: int = None,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """
        Only implemented for seq2seq models (up to 05.09 for Abs-sum & NMT)
        :return: sequences of ids of most probable tokens & scores of each sequence (sum of log probs)
        """
        if data_config is None:
            data_config = self.data_config

        if not is_tokenized:
            data = self.tokenize_data(
                data=data,
                tokenizer=self.tokenizer,
                text_name=data_config["text_name"],
                label_name=data_config["label_name"],
            )
            if "labels" in data[0]:
                data = data.remove_columns(["labels"])

        start_time = time.time()
        if getattr(self, "trainer", None) is not None:
            log.warning("Model is not fine-tuned! Be careful applying it to new data.")
        output = self._model_generate_loop(
            data, batch_size, to_eval_mode, generation_max_length, **kwargs
        )

        if calculate_time:
            self._calculate_time(start_time, phase="predict")

        if to_numpy:
            output = {k: tensor_to_numpy(v) for k, v in output.items()}

        if return_decoded_preds:
            output["predictions"] = self.tokenizer.batch_decode(
                output["sequences"], skip_special_tokens=True
            )
            if kwargs.get("num_return_sequences", 1) > 1:
                num_ret_seq = kwargs.get("num_return_sequences", 1)
                preds = output["predictions"]
                output["predictions"] = [
                    preds[i * num_ret_seq : (i + 1) * num_ret_seq]
                    for i in range(len(preds) // num_ret_seq)
                ]
                output["sequences"] = output["sequences"].reshape(
                    -1, num_ret_seq, output["sequences"].shape[1]
                )
                output["sequences_scores"] = output["sequences_scores"].reshape(
                    -1, num_ret_seq
                )

        if not return_scores:
            output.pop("sequences_scores")
        torch.cuda.empty_cache()
        gc.collect()

        return output

    def _model_generate_loop(
        self,
        data,
        batch_size=None,
        to_eval_mode=True,
        generation_max_length: int = None,
        **kwargs,
    ) -> Dict:
        """
        Returns: dict with sequences of ids and their scores
        """
        torch.cuda.empty_cache()
        if batch_size is None:
            batch_size = (
                self._batch_size_kwargs.get("generate_batch_size", None)
                or self._batch_size_kwargs.eval_batch_size
            )

        dataloader = DataLoader(
            data,
            batch_size=batch_size,
            collate_fn=DataCollatorForSeq2Seq(
                tokenizer=self.tokenizer, padding="longest"
            ),
        )

        if generation_max_length is not None:
            max_length = sequences_length = generation_max_length
        elif getattr(self, "trainer", None) is not None:
            max_length = sequences_length = self.trainer.args.generation_max_length
        else:
            max_length = None
            sequences_length = 128
        num_return_sequences = kwargs.get("num_return_sequences", 1)

        if to_eval_mode:
            self.model.eval()
        device = self.model.device
        sequences_X_shape = num_return_sequences * len(data)
        sequences = (
            torch.zeros(
                (sequences_X_shape, sequences_length), dtype=torch.int64, device=device
            )
            + self.tokenizer.pad_token_id
        )
        scores = torch.empty(sequences_X_shape, dtype=torch.float32, device=device)

        with torch.no_grad():
            start = 0
            for batch in tqdm(dataloader):
                batch = {k: v.to(device) for k, v in batch.items()}
                output = self.model.generate(
                    **batch,
                    max_length=max_length,
                    min_length=3,  # To avoid empty summaries. 3 == <BOS> + at least one token + <EOS>
                    output_scores=True,
                    return_dict_in_generate=True,
                    **kwargs,
                )
                end = start + (len(batch["input_ids"]) * num_return_sequences)
                sequences[start:end, : output.sequences.shape[1]].copy_(
                    output.sequences, non_blocking=True
                )
                scores[start:end].copy_(output.sequences_scores, non_blocking=True)
                start = end

        return {"sequences": sequences, "sequences_scores": scores}

    def _get_generation_max_length(self, data=None, data_config=None):
        if data is None:
            data = self.tokenized_dev_data
        if data is None:
            data_config = data_config if data_config is not None else self.data_config
            data = self.tokenize_data(
                tokenizer=self.tokenizer,
                data=self.dev_data,
                text_name=data_config["text_name"],
                label_name=data_config["label_name"],
            )
            self.tokenized_dev_data = data
        gen_max_length = min(
            # Using quantile instead of `max` since `max` can lead to enormously long generation
            round(np.quantile([len(x) for x in data["labels"]], 0.95) + 1),
            self.tokenizer.model_max_length,
        )
        return gen_max_length



import logging
import os
import time
from abc import abstractmethod
from pathlib import Path
from typing import Dict, Union

import mlflow
import transformers
from datasets import load_metric
from omegaconf.omegaconf import DictConfig
from torch import cuda
from datasets import Dataset
from transformers import (
    set_seed,
    EarlyStoppingCallback,
    TrainerCallback,
    PrinterCallback,
    DataCollatorForSeq2Seq
)

from .trainer_for_pseudo_labeled import TrainerForPseudoLabeled
from ...utils.general import (
    create_time_dict,
    json_dump,
    json_load,
)
from ...utils.get_train_constants import get_train_constants

log = logging.getLogger()
transformers.logging.set_verbosity_info()

CUDA_IS_AVAILABLE = (
    len(os.environ.get("CUDA_VISIBLE_DEVICES", "0")) > 0 and cuda.is_available()
)


class TransformersBaseWrapper:  # TODO: add interface
    def __init__(
        self,
        model_config,
        checkpoint,
        task: str,
        model=None,
        tokenizer=None,
        name: str = "acquisition",
        default_data_config=None,
        seed: int = 42,
        dev_data_kwargs=None,
        trainer_kwargs=None,
        batch_size_kwargs=None,
        optimizer_kwargs=None,
        scheduler_kwargs=None,
        time_dict_path: Path or str = None,
        cache_dir: Path or str = None,
        cache_model: bool = False,
        num_checkpoints_to_save: int = 1,
    ):
        self.checkpoint = checkpoint
        self.task = task
        self.name = name

        self.model = model
        self.model_config = model_config
        if tokenizer is not None:
            self.tokenizer = tokenizer
        else:
            self.set_tokenizer()

        if default_data_config is None:
            default_data_config = {
                "dataset_name": None,
                "text_name": "text",
                "label_name": "label",
            }
        self.data_config = dict(default_data_config)

        self._optimizer_kwargs = optimizer_kwargs
        self._trainer_kwargs = trainer_kwargs
        self._scheduler_kwargs = scheduler_kwargs
        self._batch_size_kwargs = batch_size_kwargs

        self.dev_data = dev_data_kwargs.pop("dev_data")
        self._dev_kwargs = DictConfig(dev_data_kwargs)
        if self.dev_data is not None and self._dev_kwargs.pop("tokenize_dev_data"):
            self.tokenized_dev_data = self.tokenize_data(
                data=self.dev_data,
                tokenizer=self.tokenizer,
                text_name=self.data_config["text_name"],
                label_name=self.data_config["label_name"],
                test_mode=False,
                **self.get_additional_tokenization_train_kwargs(),
            )
        else:
            self.tokenized_dev_data = None

        self.seed = seed
        if self._trainer_kwargs["serialization_dir"] is None:
            self._trainer_kwargs["serialization_dir"] = "./output/"
        self._num_checkpoints_to_save = num_checkpoints_to_save
        self.use_own_trainer = (
            model_config.training.get("pseudo_labeled_label_smoothing", False) != False
            or model_config.training.get("labeled_weight", 1.0) != 1.0
        )

        self.cache_dir = (
            Path(cache_dir) if cache_dir is not None else Path(f"workdir/cache_{seed}")
        )
        self.cache_dir.mkdir(exist_ok=True)
        self.cache_model = cache_model
        if time_dict_path is None:
            time_dict_path = self.cache_dir / f"time_dict_{name}.json"
            create_time_dict(time_dict_path, name)
        self.time_dict_path = time_dict_path

    def fit(
        self,
        train_data: Dataset,
        from_scratch: bool = True,
        is_tokenized: bool = False,
        data_config: Union[Dict, DictConfig] = None,
    ):

        # Here we "reverse" train and test to have train_test_split returning the data in the reversed way.
        # The idea is that for validation (i.e. development set) we want to use the instances, which would
        # best approximate the "average" instances. Since the first k instances are usually selected randomly,
        # they would be a good approximation of data in contrast to later instances, which are queried
        # according to a query strategy (usually the most uncertain ones). Hence, we would want to create
        # the development set from the randomly selected instances.
        if self._dev_kwargs["size"] > 0:
            splitted = train_data.train_test_split(
                train_size=self._dev_kwargs[
                    "size"
                ],  # this is done to have the sample in reversed way
                shuffle=self._dev_kwargs["shuffle_dev"],
                seed=self.seed,
            )
            # this is done to have the sample in reversed way
            train_data = splitted["test"]
            dev_data = splitted["train"]

            if not is_tokenized:
                data_config = (
                    data_config if data_config is not None else self.data_config
                )
                dev_data = self.tokenize_data(
                    tokenizer=self.tokenizer,
                    data=dev_data,
                    text_name=data_config["text_name"],
                    label_name=data_config["label_name"],
                    **self.get_additional_tokenization_train_kwargs(),
                )
        else:
            dev_data = self.tokenized_dev_data
            if dev_data is None and self.dev_data is not None:
                data_config = (
                    data_config if data_config is not None else self.data_config
                )
                dev_data = self.tokenize_data(
                    tokenizer=self.tokenizer,
                    data=self.dev_data,
                    text_name=data_config["text_name"],
                    label_name=data_config["label_name"],
                    **self.get_additional_tokenization_train_kwargs(),
                )
                self.tokenized_dev_data = dev_data

        log.info(f"Training dataset size: {len(train_data)}")
        log.info(
            f"Validation dataset size: {len(dev_data) if dev_data is not None else 0}"
        )

        (
            batch_size,
            num_epochs,
            steps_per_epoch,
            scheduler_warmup_steps,
        ) = get_train_constants(
            len(train_data),
            self._trainer_kwargs.num_epochs,
            self._batch_size_kwargs.batch_size,
            self._batch_size_kwargs.adjust_batch_size,
            self._batch_size_kwargs.adjust_num_epochs,
            self._batch_size_kwargs.min_num_gradient_steps,
            self._scheduler_kwargs.warmup_steps_factor,
            self._batch_size_kwargs.min_batch_size,
            self._dev_kwargs["size"],
        )

        if from_scratch or self.model is None:
            cache_dir = self.cache_dir if self.cache_model else None
            kwargs = self.get_init_model_kwargs()
            self.model, self.tokenizer = self.get_model_and_tokenizer(
                self.model_config, seed=self.seed, cache_dir=cache_dir, **kwargs
            )

        data_config = data_config if data_config is not None else self.data_config
        if not is_tokenized:
            train_data = self.tokenize_data(
                data=train_data,
                tokenizer=self.tokenizer,
                text_name=data_config["text_name"],
                label_name=data_config["label_name"],
                test_mode=False,
                **self.get_additional_tokenization_train_kwargs(),
            )

        data_collator = self.get_data_collator_class()
        if data_collator is not None:
            data_collator_kwargs = dict(tokenizer=self.tokenizer, padding="longest")
            if data_collator == DataCollatorForSeq2Seq:
                data_collator_kwargs["model"] = self.model
            data_collator = data_collator(**data_collator_kwargs)

        gradient_accumulation_steps = (
            self._trainer_kwargs.accumulation.gradient_accumulation_steps
        )
        fp16 = self._trainer_kwargs.fp16.training and CUDA_IS_AVAILABLE
        fp16_full_eval = self._trainer_kwargs.fp16.evaluation and CUDA_IS_AVAILABLE
        label_smoothing_factor = self._trainer_kwargs.get("label_smoothing_factor", 0.0)
        # Only if validation sample is "dynamic"
        load_best = self._dev_kwargs.size > 0 or self._trainer_kwargs.get(
            "load_best_at_end", False
        )
        if dev_data is None:
            logging_and_evaluation_strategy = "no"
        elif (not load_best) or (self._trainer_kwargs.evaluation_strategy != "no"):
            logging_and_evaluation_strategy = self._trainer_kwargs.evaluation_strategy
        else:
            logging_and_evaluation_strategy = "epoch"
        save_strategy = (
            self._trainer_kwargs.save_strategy
            if (self._trainer_kwargs.get("save_strategy"))
            else "no"
            if (not load_best and self.name != "successor")
            else logging_and_evaluation_strategy
        )
        save_total_limit = self._num_checkpoints_to_save
        validation_metric = "eval_" + self._trainer_kwargs.validation_metric
        other_kwargs = {}
        if self._trainer_kwargs.accumulation.eval_accumulation_steps is not None:
            other_kwargs[
                "eval_accumulation_steps"
            ] = self._trainer_kwargs.accumulation.eval_accumulation_steps
        if logging_and_evaluation_strategy == "steps":
            other_kwargs["logging_steps"] = other_kwargs[
                "save_steps"
            ] = self._trainer_kwargs.get("logging_steps", 100)

            other_kwargs["max_steps"] = self._trainer_kwargs.get(
                "max_steps", self._batch_size_kwargs.min_num_gradient_steps
            )
        log.info(f"Load best at end: {load_best}")

        training_args_class = self.get_training_args_class()
        additional_training_kwargs = self.get_additional_training_kwargs(
            train_data=train_data
        )
        other_kwargs.update(additional_training_kwargs)

        training_args = training_args_class(
            output_dir=self._trainer_kwargs["serialization_dir"]
            or "output",  # output directory
            # Batch size args
            num_train_epochs=num_epochs,  # total number of training epochs
            per_device_train_batch_size=batch_size,  # batch size per device during training
            per_device_eval_batch_size=self._batch_size_kwargs.eval_batch_size,  # batch size for evaluation
            # Optimizer args
            adafactor=self._scheduler_kwargs.use_adafactor,
            learning_rate=self._optimizer_kwargs.lr,
            weight_decay=self._optimizer_kwargs.weight_decay,  # strength of weight decay
            # Gradient args
            max_grad_norm=self._trainer_kwargs.grad_clipping,
            label_smoothing_factor=label_smoothing_factor,
            # Scheduler args
            warmup_ratio=self._scheduler_kwargs.warmup_steps_factor,
            # fp16 args
            fp16=fp16,
            fp16_full_eval=fp16_full_eval,
            # Accumulation args
            gradient_accumulation_steps=gradient_accumulation_steps,
            # Evaluation args
            metric_for_best_model=validation_metric,
            load_best_model_at_end=load_best,
            evaluation_strategy=logging_and_evaluation_strategy,  # Evaluation is done at the end of each epoch
            logging_strategy=logging_and_evaluation_strategy,
            save_strategy=save_strategy,
            save_total_limit=save_total_limit,  # limit the total amount of checkpoints. Deletes the older checkpoints
            # Disable any integrations
            report_to="none",
            # Other args
            log_level="error",
            disable_tqdm=True,
            seed=self.seed,
            **other_kwargs,
        )
        log.info(training_args)

        callbacks = [TrainingMetricsLogger()]
        if load_best:
            callbacks.append(EarlyStoppingCallback(self._trainer_kwargs.patience))

        compute_metrics = self.get_compute_metrics_fn(
            self.data_config["dataset_name"], self._trainer_kwargs.eval_metrics,
        )

        set_seed(self.seed)
        trainer_class = (
            self.get_trainer_class()
            if not self.use_own_trainer
            else TrainerForPseudoLabeled
        )
        trainer = trainer_class(
            model=self.model,
            args=training_args,
            data_collator=data_collator,
            train_dataset=train_data,
            eval_dataset=dev_data,
            callbacks=callbacks,
            compute_metrics=compute_metrics,
        )
        trainer.pop_callback(PrinterCallback)
        self.trainer = trainer

        log.info(f"Starting training...")
        start_time = time.time()

        trainer.train()

        self._calculate_time(start_time, phase="fit")

        mlflow.end_run()
        self.trainer.model.eval()
        self.best_metric = getattr(trainer.state, "best_metric", None)

    def evaluate(self, data, is_tokenized=False, data_config=None, **kwargs):
        predictions = self.get_predictions(
            data,
            is_tokenized,
            data_config,
            calculate_time=False,
            evaluate=True,
            **kwargs,
        )
        return predictions.metrics

    def optimize_hp(self, data: Dataset, hyperparameters_config):
        return

    def tokenize_data(
        self,
        data,
        tokenizer,
        text_name="text",
        label_name="labels",
        test_mode: bool = False,
        **kwargs,
    ):
        tokenize_function = self.get_tokenize_function(
            tokenizer, text_name, label_name, test_mode, **kwargs
        )
        columns_to_remove = [
            x for x in data.features.keys() if x not in ["labels", "weight"]
        ]
        # The last two arguments is a temporary fix due to datasets caching error
        return data.map(
            tokenize_function,
            batched=True,
            remove_columns=columns_to_remove,
            load_from_cache_file=False,
            cache_file_name=f"tmp.data",
        )

    def get_compute_metrics_fn(
        self, dataset_name=None, additional_metric_names=None,
    ):
        metrics_cache_dir = self.cache_dir / "metrics"
        datasets_metric_name = self.get_datasets_metric_name()
        metric = load_metric(datasets_metric_name, cache_dir=metrics_cache_dir)
        # Add metric to additional metrics in case dataset metric is available (e.g. glue-cola)
        additional_metrics = [metric]

        if dataset_name is not None:
            try:
                args = [dataset_name] if isinstance(dataset_name, str) else dataset_name
                metric = load_metric(*args, cache_dir=metrics_cache_dir)
            except:
                # Exclude the metric from additional ones
                additional_metrics = additional_metrics[:-1]

        if additional_metric_names is None:
            additional_metric_names = self.get_additional_datasets_metric_names() or []

        for add_metric_name in additional_metric_names:
            add_metric = load_metric(add_metric_name, cache_dir=metrics_cache_dir)
            additional_metrics.append(add_metric)

        kwargs = self.get_kwargs_for_compute_metrics_fn()
        return self.get_compute_metrics_function(
            metric=metric, additional_metrics=additional_metrics, **kwargs
        )

    def set_tokenizer(self):
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(self.checkpoint)

    def get_init_model_kwargs(self):
        """
        Returns a dictionary of additional arguments that will be passed to model generation
        """
        return {}

    @staticmethod
    def get_additional_datasets_metric_names():
        return []

    @staticmethod
    def get_additional_tokenization_train_kwargs():
        return {}

    @staticmethod
    def get_additional_tokenization_inference_kwargs():
        return {}

    @abstractmethod
    def get_model_and_tokenizer(
        self, model_cfg, seed: int = 42, cache_dir=None, **kwargs
    ):
        """
        Returns the model and the tokenizer
        """
        raise NotImplementedError("get_model_and_tokenizer is not implemented!")

    @abstractmethod
    def get_predictions(self, *args, **kwargs):
        """
        A function to get the model predictions
        """
        raise NotImplementedError("get_predictions is not implemented!")

    @staticmethod
    @abstractmethod
    def get_tokenize_function(
        tokenizer, text_name, label_name, test_mode: bool = False, **kwargs
    ):
        """
        Returns a function to convert a text to the list of ids (of tokens)
        """

        def tokenize_function(instances):
            raise NotImplementedError("tokenize_function is not implemented!")

        return tokenize_function

    @staticmethod
    @abstractmethod
    def get_metric_name():
        """
        Returns the name of the main metric
        """
        raise NotImplementedError("get_metric_name is not implemented!")

    @staticmethod
    @abstractmethod
    def get_datasets_metric_name():
        """
        Returns the name of the metric from `datasets` `load_metric` function
        """
        raise NotImplementedError("get_datasets_metric_name is not implemented!")

    @abstractmethod
    def get_kwargs_for_compute_metrics_fn(self):
        raise NotImplementedError(
            "get_kwargs_for_compute_metrics_fn is not implemented!"
        )

    @staticmethod
    @abstractmethod
    def get_compute_metrics_function(metric, additional_metrics, **kwargs):
        """
        Returns a function to compute metrics
        """

        def compute_metrics(eval_preds):
            raise NotImplementedError("compute_metrics is not implemented!")

        return compute_metrics

    @staticmethod
    @abstractmethod
    def get_model_class():
        """
        Returns the class of model to use
        """
        raise NotImplementedError("get_model_class is not implemented!")

    @staticmethod
    @abstractmethod
    def get_trainer_class():
        """
        Return a class for trainer
        """
        raise NotImplementedError("get_trainer_class is not implemented!")

    @staticmethod
    @abstractmethod
    def get_training_args_class():
        """
        Return a class for training arguments
        """
        raise NotImplementedError("get_training_args_class is not implemented!")

    @abstractmethod
    def get_additional_training_kwargs(self, **kwargs):
        """
        Return additional training kwargs
        """
        raise NotImplementedError("get_additional_training_kwargs is not implemented!")

    @staticmethod
    @abstractmethod
    def get_data_collator_class():
        """
        Returns a data collator that will be used in `Trainer`
        """
        raise NotImplementedError("get_data_collator_class is not implemented!")

    def enable_dropout(self):
        """
        Function to enable the dropout layers during test-time
        """
        assert self.model is not None, "Model must be set to enable dropout!"
        dropout_found = False
        for m in self.model.modules():
            if m.__class__.__name__.startswith("Dropout"):
                dropout_found = True
                m.train()

        if not dropout_found:
            self.model.train()  # e.g., for BART

    def disable_dropout(self):
        """
        Function to disable the dropout layers during test-time
        """
        assert self.model is not None, "Model must be set to disable dropout!"
        for m in self.model.modules():
            if m.__class__.__name__.startswith("Dropout"):
                m.eval()

    def _calculate_time(self, start_time, phase="fit"):
        try:
            time_work = time.time() - start_time
            time_dict = json_load(self.time_dict_path)
            time_dict[self.name + f"_{phase}"].append(time_work)
            json_dump(time_dict, self.time_dict_path)
            log.info(f"Done with the model {phase}.")
        except:
            pass


class TrainingMetricsLogger(TrainerCallback):
    def on_evaluate(self, args, state, control, metrics, **kwargs):
        # Get train loss
        log.info(f"Epoch {metrics['epoch']}\nEvaluation Metrics:")
        log.info(
            f"Train Loss: {state.log_history[-2]['loss']}"
        )  # since [-1] is the dict with evaluation metrics
        [
            log.info(f"{k.replace('_', ' ').title()}: {v}")
            for k, v in metrics.items()
            if k != "epoch"
        ]



import os

os.environ["HF_DATASETS_OFFLINE"] = "1"

from datasets import load_dataset
from sklearn.feature_extraction.text import CountVectorizer
from scipy import sparse
import numpy as np
from tqdm import tqdm
import logging
import multiprocessing as mp
from math import ceil
from pathlib import Path
import pickle

from rouge_score.tokenize import tokenize
from rouge_score.rouge_scorer import porter

logging.basicConfig(
    filename="log.txt",
    filemode="a",
    format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
    datefmt="%H:%M:%S",
    level=logging.DEBUG,
)

log = logging.getLogger()


def batch_calculate_rouges(
    i_batch, batch_size, ngramm, texts, num_ngrams, path, i_proc
):
    ngrammer = CountVectorizer(
        ngram_range=(ngramm, ngramm), tokenizer=lambda x: x.split(), lowercase=False
    )
    if batch_size < len(texts):
        batch = texts[i_batch * batch_size : (i_batch + 1) * batch_size]
        ngrammer.fit(batch)
        batch_output = ngrammer.transform(texts)

    for i_text, vector in enumerate(batch, start=i_batch * batch_size):
        log.info(f"Processing text {i_text} on CPU {i_proc}")
        result_i_text = (
            sparse.vstack([batch_output[i_text] for _ in range(batch_output.shape[0])])
            - batch_output
        )
        result_i_text[result_i_text < 0] = 0

        ngrams_i_text = batch_output[i_text].sum()
        ngrams_diff_rec = result_i_text.sum(axis=1)

        ngrams_common = ngrams_i_text - ngrams_diff_rec
        recalls_i_text = (ngrams_common / ngrams_i_text).A
        precs_i_text = (ngrams_common / num_ngrams).A
        fscores_i_text = (
            2 * recalls_i_text * precs_i_text / (precs_i_text + recalls_i_text + 1e-15)
        )

        with open(path / f"recalls_{ngramm}_{i_text}", "wb") as f:
            pickle.dump(recalls_i_text.astype(np.float16), f)
        with open(path / f"precisions_{ngramm}_{i_text}", "wb") as f:
            pickle.dump(precs_i_text.astype(np.float16), f)
        with open(path / f"fscores_{ngramm}_{i_text}", "wb") as f:
            pickle.dump(fscores_i_text.astype(np.float16), f)


if __name__ == "__main__":

    data = load_dataset("aeslc", cache_dir="data/")
    train_instances = data["train"]

    texts = [
        " ".join(tokenize(text, porter.PorterStemmer()))
        for text in tqdm(train_instances["email_body"])
    ]
    path = Path("output")
    path.mkdir(exist_ok=True)

    batch_size = 100
    if batch_size is None:
        batch_size = len(texts)

    num_procs = 16
    total_procs = ceil(len(texts) / batch_size)
    num_procs_per_core = ceil(total_procs / num_procs)

    for ngramm in range(1, 3):

        num_ngrams = (
            CountVectorizer(
                ngram_range=(ngramm, ngramm),
                tokenizer=lambda x: x.split(),
                lowercase=False,
            )
            .fit_transform(texts)
            .sum(axis=1)
        )
        processes = []

        for i_core_proc in range(num_procs_per_core):
            for i in tqdm(range(num_procs)):
                idx_batch = i + i_core_proc * num_procs
                if idx_batch * batch_size >= len(texts):
                    continue

                proc = mp.Process(
                    target=batch_calculate_rouges,
                    args=(idx_batch, batch_size, ngramm, texts, num_ngrams, path, i),
                )
                proc.start()
                processes.append(proc)
            for proc in processes:
                proc.join()



from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import get_query_idx_for_selecting_by_number_of_tokens
from ..utils.transformers_dataset import TransformersDataset


def random_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    select_by_number_of_tokens: bool = False,
    **kwargs,
):
    if select_by_number_of_tokens:
        sorted_idx = np.arange(len(X_pool))
        np.random.shuffle(sorted_idx)
        query_idx = get_query_idx_for_selecting_by_number_of_tokens(
            X_pool, sorted_idx, n_instances
        )
    else:
        query_idx = np.random.choice(range(len(X_pool)), n_instances, replace=False)

    query = X_pool.select(query_idx)
    # Uncertainty estimates are not defined with random sampling
    uncertainty_estimates = np.zeros(len(X_pool))

    return query_idx, query, uncertainty_estimates



import logging
from typing import Union

import numpy as np
import torch
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import take_idx
from .strategy_utils.batchbald.batchbald import get_batchbald_batch
from .strategy_utils.batchbald.consistent_dropout import make_dropouts_consistent
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def batchbald(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **bald_kwargs,
):
    """
    Jointly score points by estimating the mutual information between a joint
    of multiple data points and the model parameters. https://arxiv.org/abs/1906.08158
    """
    mc_iterations = bald_kwargs.get("mc_iterations", 10)
    max_num_samples = bald_kwargs.get("max_num_samples", int(1e4))
    # requires enormous amount of memory
    device = "cpu"  # list(model.model.parameters())[0].device

    # Make dropout consistent inside huggingface model
    make_dropouts_consistent(model.model)

    if bald_kwargs.get("only_head_dropout", False):
        raise NotImplementedError
        # log_probas_N_K_C = model.predict_proba(X_pool, mc_dropout=True, mc_iterations=mc_iterations)
    else:
        probas = []
        for _ in range(mc_iterations):
            # Reset masks
            model.enable_dropout()
            model.disable_dropout()
            probas_iter = model.predict_proba(
                X_pool, use_predict_loop=True, to_numpy=False, to_eval_mode=False
            )
            probas.append(probas_iter)

        log_probas_N_K_C = torch.stack(probas, -2).log().to(device)

    query_idx, uncertainty_estimates = get_batchbald_batch(
        log_probas_N_K_C, n_instances, max_num_samples, device=device
    )
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates



import copy
import logging
import time

import numpy as np
from scipy.stats import rankdata

from .mahalanobis_distance import (
    mahalanobis_distance,
    mahalanobis_distance_relative,
    mahalanobis_distance_marginal,
    compute_centroids,
    compute_covariance,
)
from .utils_heads import (
    ElectraClassificationHeadIdentityPooler,
    BertClassificationHeadIdentityPooler,
    ElectraNERHeadIdentityPooler,
)
from .utils_inference import is_custom_head, unpad_features, pad_scores

log = logging.getLogger()


def entropy(x):
    return np.sum(-x * np.log(np.clip(x, 1e-8, 1)), axis=-1)


class UeEstimatorHybrid:
    def __init__(self, cls, ue_args, config, train_dataset):
        self.cls = cls
        self.ue_args = ue_args
        self.config = config
        self.train_dataset = train_dataset

    def __call__(self, X):
        return self._predict_with_fitted_cov(X)

    def fit_ue(self, X, y=None, X_test=None):
        cls = self.cls
        model = self.cls._auto_model

        log.info(
            "****************Start fitting covariance and centroids **************"
        )

        if y is None:
            y = self._extract_labels(X)

        self._replace_model_head()
        X_features = self._extract_features(X)

        self.class_cond_centroids = self._fit_centroids(X_features, y)
        self.class_cond_covarince = self._fit_covariance(X_features, y)

        self.fit_all_md_versions = (
            "fit_all_md_versions" in self.ue_args.keys()
            and self.ue_args["fit_all_md_versions"]
        )
        if self.fit_all_md_versions:
            self.train_centroid = self._fit_centroids(X_features, y, class_cond=False)
            self.train_covarince = self._fit_covariance(X_features, y, class_cond=False)

        log.info("**************Done.**********************")

    def _fit_covariance(self, X, y, class_cond=True):
        if class_cond:
            return compute_covariance(self.class_cond_centroids, X, y, class_cond)
        return compute_covariance(self.train_centroid, X, y, class_cond)

    def _fit_centroids(self, X, y, class_cond=True):
        return compute_centroids(X, y, class_cond)

    def _replace_model_head(self):
        log.info("Change classifier to Identity Pooler")
        cls = self.cls
        model = self.cls._auto_model
        self.old_head = copy.deepcopy(model.classifier)

        if is_custom_head(model):
            model.classifier = ElectraClassificationHeadIdentityPooler(model.classifier)
        else:
            model.classifier = BertClassificationHeadIdentityPooler(model.classifier)

    def _return_head(self):
        self.cls._auto_model.classifier = self.old_head
        log.info("Change Identity Pooler to classifier")

    def _extract_labels(self, X):
        return np.asarray([example["labels"] for example in X])

    def _extract_features(self, X):
        cls = self.cls
        model = self.cls._auto_model
        X = X.remove_columns("labels")

        X_features = cls.predict(X, apply_softmax=False, return_preds=True)[0]
        return X_features

    def _predict_with_fitted_cov(self, X):
        cls = self.cls
        model = self.cls._auto_model

        log.info(
            "****************Compute MD with fitted covariance and centroids **************"
        )

        self._return_head()
        probs = cls.predict(X, apply_softmax=True, return_preds=False)[0]
        unc_entropy = entropy(probs)
        self._replace_model_head()
        self.md_threshold = (
            self.ue_args["md_threshold"]
            if "md_threshold" in self.ue_args.keys()
            else 0.2
        )

        start = time.time()
        X_features = self._extract_features(X)

        end = time.time()

        eval_results = {}

        md, inf_time = mahalanobis_distance(
            None, None, X_features, self.class_cond_centroids, self.class_cond_covarince
        )

        sum_inf_time = inf_time + (end - start)

        n_preds = len(X)
        n_lowest = int(n_preds * self.md_threshold)

        md_rank = rankdata(md)
        unc_entropy_rank = rankdata(unc_entropy[md_rank < n_lowest])
        md_rank[md_rank < n_lowest] = unc_entropy_rank

        eval_results["mahalanobis_distance"] = md_rank.tolist()
        eval_results["ue_time"] = sum_inf_time
        log.info(f"UE time: {sum_inf_time}")

        if self.fit_all_md_versions:
            md_relative = mahalanobis_distance_relative(
                None, None, X_features, self.train_centroid, self.train_covarince
            )

            md_marginal = mahalanobis_distance_marginal(
                None,
                None,
                X_features,
                self.class_cond_centroids,
                self.class_cond_covarince,
                self.train_centroid,
                self.train_covarince,
            )

            eval_results["mahalanobis_distance_relative"] = md_relative.tolist()
            eval_results["mahalanobis_distance_marginal"] = md_marginal.tolist()

        log.info("**************Done.**********************")
        return eval_results


class UeEstimatorMahalanobisNer:
    def __init__(self, cls, ue_args, config, train_dataset):
        self.cls = cls
        self.ue_args = ue_args
        self.config = config
        self.train_dataset = train_dataset

    def __call__(self, X, y):
        return self._predict_with_fitted_cov(X, y)

    def fit_ue(self, X, y=None, X_test=None):
        cls = self.cls
        model = self.cls._auto_model

        log.info(
            "****************Start fitting covariance and centroids **************"
        )

        if y is None:
            y, y_shape = self._extract_labels(X)

        self._replace_model_head()
        X_features = self._extract_features(X)
        X_features, y = unpad_features(X_features, y)

        self.class_cond_centroids = self._fit_centroids(X_features, y)
        self.class_cond_covarince = self._fit_covariance(X_features, y)

        self.fit_all_md_versions = (
            "fit_all_md_versions" in self.ue_args.keys()
            and self.ue_args["fit_all_md_versions"]
        )

        if self.fit_all_md_versions:
            self.train_centroid = self._fit_centroids(X_features, y, class_cond=False)
            self.train_covarince = self._fit_covariance(X_features, y, class_cond=False)

        log.info("**************Done.**********************")

    def _fit_covariance(self, X, y, class_cond=True):
        if class_cond:
            return compute_covariance(self.class_cond_centroids, X, y, class_cond)
        return compute_covariance(self.train_centroid, X, y, class_cond)

    def _fit_centroids(self, X, y, class_cond=True):
        return compute_centroids(X, y, class_cond)

    def _replace_model_head(self):
        log.info("Change classifier to Identity Pooler")
        cls = self.cls
        model = self.cls._auto_model

        if is_custom_head(model):
            model.classifier = ElectraNERHeadIdentityPooler(model.classifier)
        else:
            model.classifier = BertClassificationHeadIdentityPooler(model.classifier)

    def _extract_labels(self, X):
        y = np.asarray([example["labels"] for example in X])
        y_shape = y.shape

        return y.reshape(-1), y_shape

    def _extract_features(self, X):
        cls = self.cls
        model = self.cls._auto_model

        try:
            X = X.remove_columns("labels")
        except:
            X.dataset = X.dataset.remove_columns("labels")

        X_features = cls.predict(X, apply_softmax=False, return_preds=False)[0]
        X_features = X_features.reshape(-1, X_features.shape[-1])

        return X_features

    def _predict_with_fitted_cov(self, X, y):
        cls = self.cls
        model = self.cls._auto_model

        log.info(
            "****************Compute MD with fitted covariance and centroids **************"
        )

        start = time.time()

        y_pad, y_shape = self._extract_labels(X)
        X_features = self._extract_features(X)
        X_features, y = unpad_features(X_features, y_pad)

        end = time.time()

        eval_results = {}

        md, inf_time = mahalanobis_distance(
            None, None, X_features, self.class_cond_centroids, self.class_cond_covarince
        )

        md = pad_scores(md, np.asarray(y_pad).reshape(y_shape), y_pad)

        sum_inf_time = inf_time + (end - start)
        eval_results["mahalanobis_distance"] = md.tolist()
        eval_results["ue_time"] = sum_inf_time
        log.info(f"UE time: {sum_inf_time}")

        if self.fit_all_md_versions:
            md_relative = mahalanobis_distance_relative(
                None, None, X_features, self.train_centroid, self.train_covarince
            )
            md_relative = pad_scores(
                md_relative, np.asarray(y_pad).reshape(y_shape), y_pad
            )

            md_marginal = mahalanobis_distance_marginal(
                None,
                None,
                X_features,
                self.class_cond_centroids,
                self.class_cond_covarince,
                self.train_centroid,
                self.train_covarince,
            )
            md_relative = pad_scores(
                md_relative, np.asarray(y_pad).reshape(y_shape), y_pad
            )

            eval_results["mahalanobis_distance_relative"] = md_relative.tolist()
            eval_results["mahalanobis_distance_marginal"] = md_marginal.tolist()

        log.info("**************Done.**********************")
        return eval_results



import numpy as np
import pandas as pd
from ipywidgets import Button, VBox, HBox, HTML, Box, Layout, Label, Text, ToggleButtons
from traitlets import Int, Instance

from .visualizers import TextAreaVisualizer


class AnnotatorWidget(Box):
    """The widget for Jupyter that implements example annotator.

    The widget can be used without active learning.

    """

    _max_examples = Int(5, allow_none=False).tag(sync=True)
    _current_position = Int(0, allow_none=False).tag(sync=True)
    _dataframe = Instance(klass=pd.DataFrame)

    _example_layout = Instance(klass=Layout)
    _example_layout = Layout(
        width="99%",
        border="solid 2px",
        margin="3px",
        align_items="stretch",
        padding="2px",
    )

    ANNOT_DONT_KNOW = "Don't know"

    def __init__(
        self,
        dataframe,
        answers=None,
        max_examples=5,
        current_position=0,
        visualize_columns=[],
        drop_labels=[],
        y_labels={"True": True, "False": False},
        visualizer=None,
        display_feature_table=True,
        *args,
        **kwargs
    ):
        """Annotator constructor.

        Args:
            dataframe (pandas.DataFrame):  the dataframe that contains data for examples visualizatin.
            answers (numpy.array): the array that contains known answers
                (that will not be marked as Don't Know check button).
            max_examples (int): maximum number of examples per page.
            current_position (int): index of position, from which iteration should start.
            textual_labels (list): list of string labels that will be visualized with VisualizerTextArea.
            drop_labels (list): list of string labels that will be dropped from visualization via table.
            y_labels (dict): dict {<y_textual_label> : <y_value>}.
            visualizer (object): visualizer for X_helper representation. The default is None. If None the widget
                will invoke VisualizerTextArea by deafult.

        """
        super(Box, self).__init__(*args, **kwargs)

        self._y_labels = None
        if y_labels is not None:
            self._y_labels = y_labels
            self._y_labels[self.ANNOT_DONT_KNOW] = None
            self._y_labels_reversed = {v: k for k, v in self._y_labels.items()}

        self._display_feature_table = display_feature_table

        self._dataframe = dataframe
        self._current_position = min(current_position, self._dataframe.shape[0] - 1)
        self._max_examples = max_examples
        self._answers = (
            answers
            if answers is not None
            else np.array([None] * self._dataframe.shape[0])
        )
        assert (
            self._answers.shape[0] == self._dataframe.shape[0]
        ), "The length of dataframe should match the length of numpy.array with answers."

        self._drop_labels = drop_labels + visualize_columns
        self._visualizer = visualizer
        if self._visualizer is None:
            self._visualizer = TextAreaVisualizer(visualize_columns)
        self._visualizer.init(dataframe=self._dataframe, answers=self.get_answers())

        self._draw()
        self.observe(self._draw, names="_current_position")

    def get_answers(self):
        """Returns numpy.array with answers."""
        return self._answers

    def get_dataframe(self):
        """Returns pandas.DataFrame with feature values."""
        return self._dataframe

    def _click_prev(self, button):
        self._current_position = max(self._current_position - self._max_examples, 0)

    def _click_next(self, button):
        self._current_position = min(
            self._current_position + self._max_examples, self._dataframe.shape[0]
        )

    def _int_text_value_changed(self, wdg):
        try:
            new_value = int(wdg.value)
        except ValueError:
            return

        if new_value < 0:
            new_value = self._dataframe.shape[0] + new_value
            if new_value < 0:
                return

        if new_value >= self._dataframe.shape[0]:
            return

        self._current_position = new_value

    def _make_controls(self):
        controls = HBox(
            children=[
                Button(description="Prev"),
                Button(description="Next"),
                Text(value=str(self._current_position), layout=Layout(width="80px")),
                Label(value="out of", layout=Layout(width="35px")),
                Text(
                    value=str(self._dataframe.shape[0]),
                    disabled=True,
                    layout=Layout(width="80px"),
                ),
            ]
        )
        controls.children[0].on_click(self._click_prev)
        controls.children[1].on_click(self._click_next)
        controls.children[2].observe(self._int_text_value_changed, "")

        return controls

    def _annotate(self, num, change):
        ch = change["new"]
        if ch == self.ANNOT_DONT_KNOW:
            self._answers[num] = None
        else:
            self._answers[num] = self._y_labels[ch]

    def _answer_to_label(self, answer):
        return self._y_labels_reversed[answer]

    def _draw(self, change=None):
        self._table = VBox(layout=Layout(width="100%"))
        self._table.children += (self._make_controls(),)

        last_element = min(
            self._current_position + self._max_examples, self._dataframe.shape[0]
        )

        for i in range(self._current_position, last_element):
            data_row = VBox(layout=self._example_layout)

            if self._display_feature_table:
                elem = self._dataframe.iloc[i].drop(self._drop_labels)
                data_row.children += (
                    HTML(
                        value=pd.DataFrame(
                            [elem.values],
                            columns=elem.index,
                            index=[self._dataframe.index[i]],
                        ).to_html(classes=["table", "table-striped"])
                    ),
                )

            data_row.children += self._visualizer(i)

            if self._y_labels is not None:
                data_row.children += (
                    ToggleButtons(
                        options=(
                            [self.ANNOT_DONT_KNOW]
                            + [
                                k
                                for k in self._y_labels.keys()
                                if k != self.ANNOT_DONT_KNOW
                            ]
                        ),
                        value=self._answer_to_label(self._answers[i]),
                        description="Your annotation:",
                        disabled=False,
                    ),
                )
                data_row.children[-1].observe(
                    lambda tgl_bt, num=i: self._annotate(num, tgl_bt), names="value"
                )

            self._table.children += (data_row,)

        self._table.children += (self._make_controls(),)
        self.children = (self._table,)



from torch.nn.functional import softmax
from transformers import Trainer


class TrainerForPseudoLabeled(Trainer):
    def compute_loss(self, model, inputs, return_outputs=False):
        """
        Overrided for label smoothing for pseudo-labeled data.
        How the loss is computed by Trainer. By default, all models return the loss in the first element.
        """
        labels = inputs.pop(
            "labels"
        )  # we assume that labels are presented as vectors regardless of whether they are smoothed
        weight = None
        if "weight" in inputs:
            weight = inputs.pop("weight")
        outputs = model(**inputs)
        if self.args.past_index >= 0:
            self._past = outputs[self.args.past_index]

        logits = outputs.logits
        if labels.shape != logits.shape:
            labels = labels.unsqueeze(1)
            losses = -softmax(logits, dim=-1).log().gather(1, labels)
        else:
            losses = -softmax(logits, dim=-1).log() * labels

        if weight is not None:
            loss = (losses * weight).sum() / len(labels)
        else:
            loss = losses.sum() / len(labels)

        return (loss, outputs) if return_outputs else loss



import logging

import fastcluster
import numpy as np
from scipy.cluster.hierarchy import fcluster
from scipy.spatial.distance import cdist

from .get_embeddings import get_embeddings

log = logging.getLogger()


def hierarchial_ac(classifier, X_pool, X_train, config):
    """Wrapper for AgglomerativeClustering from scipy and fastcluster."""
    # get data embeddings
    kwargs = dict(
        # General
        prepare_model=True,
        use_activation=False,
        use_spectralnorm=False,
        data_is_tokenized=False,
        batch_size=classifier._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        # Tokenization
        tokenizer=classifier.tokenizer,
        task=classifier.task,
        text_name=classifier.data_config["text_name"],
        label_name=classifier.data_config["label_name"],
    )
    distance = config.al.strategy_kwargs.get("distance", "cosine")
    fast = config.al.strategy_kwargs.get("fast", False)
    if fast:
        # in this case, we calc clusters only for part of data, and after calc centroids for these clusters
        # train_data = X_pool.select(np.random.choice(np.arange(len(X_pool)), int(0.1 * len(X_pool))))
        train_data = X_train
        embeddings = get_embeddings(classifier.model, train_data, **kwargs)
        n_clusters = config.al.strategy_kwargs.get("n_clusters", 12000)
        log.info("Started clustering")
        linkage_matrix = fastcluster.linkage(
            embeddings, method="average", metric=distance
        )
        clusters = fcluster(linkage_matrix, t=n_clusters, criterion="maxclust")
        log.info("Done with clustering on train")
        centroids = {}
        unique_clusters = np.unique(clusters)
        for cluster in unique_clusters:
            # calc centroid for this cluster
            centroids[cluster] = np.average(
                embeddings[np.where(clusters == cluster)], axis=0
            )
        # after assign all embeddings to nearest cluster
        embeddings = get_embeddings(classifier.model, X_pool, **kwargs)
        centroids_values = np.array(list(centroids.values()))
        distances = cdist(centroids_values, embeddings, metric=distance)
        cluster_ids = np.argmin(distances, axis=0)
        clusters = np.array(list(centroids.keys()))[cluster_ids.astype(int)].astype(int)
        del embeddings, distances
        log.info("Done with clustering")
    else:
        train_data = X_pool
        embeddings = get_embeddings(classifier.model, train_data, **kwargs)
        n_clusters = config.al.strategy_kwargs.get("n_clusters", 12000)
        log.info("Started clustering")
        linkage_matrix = fastcluster.linkage(
            embeddings, method="average", metric=distance
        )
        del embeddings
        clusters = fcluster(linkage_matrix, t=n_clusters, criterion="maxclust")
        log.info("Done with clustering")
    return clusters


"""
def cluster_margin(classifier, X_pool, n_instances, **kwargs):
    instances_multiplier = kwargs["instances_multiplier"]
    random_query = kwargs["random_query"]
    probas = classifier.predict_proba(X_pool)
    # reuse code from margin sampling
    # To get second max probas, need to sort the array since `.sort` modifies the array
    probas.sort(axis=1)
    max_probas = probas[:, -1]
    second_max_probas = probas[:, -2]
    uncertainty_estimates = 1 + second_max_probas - max_probas
    argsort = np.argsort(-uncertainty_estimates)
    # we have to choose n_instances with round-robin algorithm, so for this we
    # firstly choose 2 * n_instances
    query_idx = argsort[:int(instances_multiplier * n_instances)]
    cluster_labels = np.array(X_pool.clusters)[query_idx]
    # X_pool either transformer dataset or np array
    query_idx, samples_idx = choose_samples(cluster_labels, query_idx, n_instances, random_query)
    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates


def choose_samples(cluster_labels, query_idx, n_instances, random_query):
    # assume that we added cluster labels as coulmn in dataset
    # so we get samples, scores and clusters indices
    cluster_sizes = Counter(cluster_labels)
    new_query_idx = []
    samples_idx = []
    # split all query_idx to array by cluster label
    query_idx_by_clusters = {idx: list(query_idx[np.where(cluster_labels == idx)]) for idx in list(cluster_sizes.keys())}

    # try faster approach - collect all indices that we will sample
    samples_per_cluster = []
    avg_sample_per_cluster = np.ceil(n_instances / len(cluster_sizes))

    sorted_cluster = [el[1] for el in sorted([(v, k) for k, v in cluster_sizes.items()])]
    # else find smaller cluster, and sample its size, after subtract n_clusters * min_size and do same
    curr_idx = 0
    while sum(cluster_sizes.values()) > 0 and len(new_query_idx) < n_instances:
        # sample data from each cluster
        # sorted_cluster - array with cluster numbers in ascending order by size
        # cluster_labels - dict with cluster sizes as values
        curr_cluster = sorted_cluster[curr_idx]
        if cluster_sizes[curr_cluster] == 0:
            curr_idx = (curr_idx + 1) % len(sorted_cluster)
            continue
        # randomly sample from data with curr_cluster labels
        if random_query:
            sample_idx = random.choice(np.arange(len(query_idx_by_clusters[curr_cluster])))
        else:
            sample_idx = 0
        samples_idx.append(sample_idx)
        new_query_idx.append(query_idx_by_clusters[curr_cluster][sample_idx])
        # and remove this sample from data
        query_idx_by_clusters[curr_cluster] = np.delete(query_idx_by_clusters[curr_cluster], sample_idx)
        # after subtract 1 from this cluster
        cluster_sizes[curr_cluster] -= 1
        curr_idx = (curr_idx + 1) % len(sorted_cluster)
    return np.array(new_query_idx), samples_idx
"""



import numpy as np
import torch
from tqdm import tqdm

DOUBLE_INFO = torch.finfo(torch.double)
JITTERS = [10 ** exp for exp in range(-15, 0, 1)]


def _compute_centroid(train_features, train_labels, label, zero_vector=None):
    label_features = train_features[train_labels == label]
    if len(label_features):
        return label_features.mean(dim=0), False
    return zero_vector, True


def compute_centroids(train_features, train_labels, num_labels=None):

    labels = (
        np.sort(np.unique(train_labels))
        if num_labels is None
        else np.arange(num_labels)
    )
    device = train_features.device
    centroids = torch.empty(
        len(labels), train_features.shape[1], dtype=torch.float32, device=device
    )
    centroids_mask = torch.empty(len(labels), dtype=torch.bool, device="cpu")
    zero_vector = torch.zeros(train_features.shape[1], device=device)

    for i, label in enumerate(labels):
        centroid, centroid_mask = _compute_centroid(
            train_features, train_labels, label, zero_vector
        )
        centroids[i].copy_(centroid, non_blocking=True)
        centroids_mask[i] = centroid_mask

    return centroids, centroids_mask


def compute_inv_covariance(centroids, train_features, train_labels, jitters=None):
    if jitters is None:
        jitters = JITTERS
    jitter = 0
    jitter_eps = None

    cov = torch.zeros(
        centroids.shape[1], centroids.shape[1], device=centroids.device
    ).float()
    for c, mu_c in tqdm(enumerate(centroids)):
        for x in train_features[train_labels == c]:
            d = (x - mu_c).unsqueeze(1)
            cov += d @ d.T
    cov_scaled = cov / (train_features.shape[0] - 1)

    for i, jitter_eps in enumerate(jitters):
        jitter = jitter_eps * torch.eye(cov_scaled.shape[1], device=cov_scaled.device,)
        cov_scaled_update = cov_scaled + jitter
        eigenvalues = torch.symeig(cov_scaled_update).eigenvalues
        if (eigenvalues >= 0).all():
            break
    cov_scaled = cov_scaled + jitter
    cov_inv = torch.inverse(cov_scaled.to(torch.float64)).float()
    return cov_inv, jitter_eps


def mahalanobis_distance(train_features, train_labels, eval_features):
    centroids, centroids_mask = compute_centroids(train_features, train_labels)
    sigma = compute_inv_covariance(centroids, train_features, train_labels)
    diff = eval_features[:, None, :] - centroids[None, :, :]
    sigma_inv = np.linalg.pinv(sigma)

    dists = np.matmul(np.matmul(diff, sigma_inv), diff.transpose(0, 2, 1))
    dists = np.asarray([np.diag(dist) for dist in dists])
    return np.min(dists, axis=1)


def mahalanobis_distance_with_known_centroids_sigma_inv(
    centroids, centroids_mask, sigma_inv, eval_features
):
    diff = eval_features.unsqueeze(1) - centroids.unsqueeze(
        0
    )  # bs (b), num_labels (c / s), dim (d / a)
    dists = torch.sqrt(torch.einsum("bcd,da,bsa->bcs", diff, sigma_inv, diff))
    device = dists.device
    dists = torch.stack([torch.diag(dist).cpu() for dist in dists], dim=0)
    dists = dists.masked_fill_(centroids_mask, float("inf")).to(device)
    return dists  # np.min(dists, axis=1)



import gc
import logging
import pickle
import random
import time
from collections import Counter
from math import ceil
from pathlib import Path
from typing import List, Union, Tuple

import numpy as np
import torch
from datasets import load_metric, concatenate_datasets
from datasets.arrow_dataset import Dataset as ArrowDataset
from nltk.tokenize import sent_tokenize, word_tokenize
from nltk.translate.bleu_score import corpus_bleu
from omegaconf.dictconfig import DictConfig
from requests.models import HTTPError
from sklearn.cluster import KMeans
from sklearn.neighbors import KNeighborsClassifier
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import (
    AutoModel,
    AutoTokenizer,
    AutoModelForMaskedLM,
)
from transformers.data.data_collator import torch_default_data_collator

from .strategy_utils import (
    compute_centroids,
    compute_inv_covariance,
    compute_inv_covariance_v2,
    mahalanobis_distance_with_known_centroids_sigma_inv,
    mahalanobis_distance_with_known_centroids_sigma_inv_v2,
    gmm_fit,
    get_gmm_log_probs,
    compute_density,
)
from .strategy_utils.gmm import class_probs, gmm_evaluate, gmm_fit_, compute_density_
from ..utils.general import json_dump, json_load
from ..utils.get_embeddings import get_embeddings
from ..utils.transformers_dataset import TransformersDataset
from ..model_wrappers.transformers import WrapperCls, WrapperNer, WrapperAts, WrapperNmt

log = logging.getLogger()
WRAPPERS = {"cls": WrapperCls, "ner": WrapperNer, "ats": WrapperAts, "nmt": WrapperNmt}


def get_query_idx_for_selecting_by_number_of_tokens(
    data: Union[
        List[Union[str, List[str]]], np.ndarray, ArrowDataset, TransformersDataset,
    ],
    sorted_idx: List[int] or np.ndarray[int],
    num_to_select: int,
    tokens_column_name: str or None = None,
    include_last: bool = True,
):
    """
    Function to get the id of the last selected sample `id_bound`, so that sorted_data[:id_bound] will be selected
    """
    sorted_data = (
        data[sorted_idx]
        if isinstance(data, np.ndarray)
        else [data[i] for i in sorted_idx]
        if isinstance(data, list)
        else data.select(sorted_idx)
    )

    if tokens_column_name is None:
        tokens_column_name = "tokens"

    if isinstance(sorted_data[0], str):
        sample_num_tokens = [len(x.split()) for x in sorted_data]
    elif isinstance(sorted_data[0], list):
        sample_num_tokens = [len(x) for x in sorted_data]
    else:
        sample_num_tokens = [len(x) for x in sorted_data[tokens_column_name]]

    cumsum_num_tokens = np.cumsum(sample_num_tokens)
    # get index of last id to select
    last_id = np.argwhere(cumsum_num_tokens > num_to_select).ravel().min() + int(
        include_last
    )
    query_idx = sorted_idx[:last_id]
    return query_idx


def calculate_mnlp_score(probas) -> np.ndarray:
    return np.array([-np.sum(np.log(np.max(i, axis=1))) / len(i) for i in probas])


def calculate_mnlp_scores_empty_penalty(probas) -> np.ndarray:
    positiveness = np.max(np.argmax(probas, axis=-1), axis=-1) != 0
    scores = np.array([-np.sum(np.log(np.max(i, axis=1))) / len(i) for i in probas])
    return np.where(positiveness, scores, -np.inf)


def probability_variance_ner(sampled_probabilities, mean_probabilities=None):
    if mean_probabilities is None:
        mean_probabilities = np.mean(sampled_probabilities, axis=1)

    mean_probabilities = np.expand_dims(mean_probabilities, axis=1)
    tmp = sampled_probabilities - mean_probabilities
    tmp = np.array([i ** 2 for i in tmp])
    tmp = np.array([np.mean(i) for i in tmp])
    return np.array([np.sum(i) for i in tmp])


def mean_entropy_ner(sampled_probabilities):
    sum_m = np.mean(sampled_probabilities, axis=1)
    H = np.array([-np.sum(i * np.log(np.clip(i, 1e-8, 1)), axis=-1) for i in sum_m])
    return np.array([np.mean(i) for i in H])


def calculate_alps_scores(
    model_wr,
    dataloader_or_data: Union[DataLoader, ArrowDataset, TransformersDataset],
    data_is_tokenized=False,
    batch_size: int = 16,
    **tokenization_kwargs,
):

    # model_name = get_name(model)
    device = next(model_wr.model.parameters()).device
    if device.type == "cuda":
        model_wr.model.cpu()
        torch.cuda.empty_cache()
    model = AutoModelForMaskedLM.from_pretrained(model_wr.model_config.checkpoint).to(
        device
    )

    """Return scores (or vectors) for data [batch] given the active learning method"""
    model.eval()

    model_wr.tokenizer.model_max_length = 256
    wrapper = WRAPPERS[tokenization_kwargs.pop("task", "cls")]
    if not isinstance(dataloader_or_data, DataLoader):
        if not data_is_tokenized:
            if "label_name" not in tokenization_kwargs:
                tokenization_kwargs["test_mode"] = True
            tokenize_function = wrapper.get_tokenize_function(**tokenization_kwargs,)
            columns_to_remove = [
                x
                for x in dataloader_or_data.features.keys()
                if x not in ["labels", "weight"]
            ]
            # The last two arguments is a temporary fix for Kristofari
            dataloader_or_data = dataloader_or_data.map(
                tokenize_function,
                batched=True,
                remove_columns=columns_to_remove,
                load_from_cache_file=False,
                cache_file_name=f"tmp.data",
            )
            if isinstance(wrapper, WrapperNer):
                idx_first_bpe = dataloader_or_data["idx_first_bpe"]
                dataloader_or_data = dataloader_or_data.remove_columns("idx_first_bpe")
        dataloader_or_data = DataLoader(
            dataloader_or_data,
            shuffle=False,
            batch_size=batch_size,
            collate_fn=wrapper.get_data_collator_class()(
                tokenizer=tokenization_kwargs["tokenizer"],
                padding="max_length",
                max_length=256,
            ),
            pin_memory=(str(device).startswith("cuda")),
        )

    num_obs = len(dataloader_or_data.dataset)

    start = 0
    all_scores_or_vectors = None
    for batch in tqdm(dataloader_or_data, desc="Embeddings created"):
        batch = {k: v.to(device) for k, v in batch.items()}

        inputs = {}
        # mask_tokens() requires CPU input_ids
        input_ids_cpu = batch["input_ids"].cpu().clone()
        input_ids_mask, labels = mask_tokens(
            input_ids_cpu, tokenization_kwargs["tokenizer"]
        )
        input_ids = batch["input_ids"]
        input_ids = input_ids.to(device)
        labels = labels.to(device)
        inputs["input_ids"] = input_ids
        inputs["labels"] = labels

        inputs["attention_mask"] = batch["attention_mask"]
        # if args.model_type != "distilbert":
        # if args.model_type in ["bert", "xlnet", "albert"] else None
        # XLM, DistilBERT, RoBERTa, and XLM-RoBERTa don't use segment_ids
        # """Obtain masked language modeling loss from [model] for tokens in [inputs].
        # Should return batch_size X seq_length tensor."""
        logits = model(**inputs)[1]
        labels = inputs["labels"]
        batch_size, seq_length, vocab_size = logits.size()
        loss_fct = nn.CrossEntropyLoss(reduction="none")
        loss_batched = loss_fct(logits.view(-1, vocab_size), labels.view(-1))
        loss = loss_batched.view(batch_size, seq_length)

        if all_scores_or_vectors is None:
            all_scores_or_vectors = loss.detach().cpu().numpy()
        else:
            all_scores_or_vectors = np.append(
                all_scores_or_vectors, loss.detach().cpu().numpy(), axis=0
            )
    all_scores_or_vectors = torch.tensor(all_scores_or_vectors)

    # Return the original model to cuda
    if device.type == "cuda":
        torch.cuda.empty_cache()
        model_wr.model.to(device)

    return all_scores_or_vectors


def mask_tokens(inputs, tokenizer):
    """Prepare masked tokens inputs/labels for masked language modeling: 80% MASK, 10% random, 10% original."""

    if tokenizer.mask_token is None:
        raise ValueError(
            "This tokenizer does not have a mask token which is necessary for masked language modeling. Remove the --mlm flag if you want to use this tokenizer."
        )

    labels = inputs.clone()
    # We sample a few tokens in each sequence for masked-LM training (with probability args.mlm_probability defaults to 0.15 in Bert/RoBERTa)
    probability_matrix = torch.full(labels.shape, 0.15)
    special_tokens_mask = [
        tokenizer.get_special_tokens_mask(val, already_has_special_tokens=True)
        for val in labels.tolist()
    ]
    probability_matrix.masked_fill_(
        torch.tensor(special_tokens_mask, dtype=torch.bool), value=0.0
    )
    if tokenizer._pad_token is not None:
        padding_mask = labels.eq(tokenizer.pad_token_id)
        probability_matrix.masked_fill_(padding_mask, value=0.0)
    masked_indices = torch.bernoulli(probability_matrix).bool()
    labels[~masked_indices] = -100  # We only compute loss on masked tokens

    # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
    indices_replaced = (
        torch.bernoulli(torch.full(labels.shape, 0.8)).bool() & masked_indices
    )
    inputs[indices_replaced] = tokenizer.convert_tokens_to_ids(tokenizer.mask_token)

    # 10% of the time, we replace masked input tokens with random word
    indices_random = (
        torch.bernoulli(torch.full(labels.shape, 0.5)).bool()
        & masked_indices
        & ~indices_replaced
    )
    random_words = torch.randint(len(tokenizer), labels.shape, dtype=torch.long)
    inputs[indices_random] = random_words[indices_random]

    # The rest of the time (10% of the time) we keep the masked input tokens unchanged
    return inputs, labels


def entropy(x):
    return np.sum(-x * np.log(np.clip(x, 1e-8, 1)), axis=-1)


def mean_entropy(sampled_probabilities):
    return entropy(np.mean(sampled_probabilities, axis=1))


def calculate_bald_score_cls(sampled_probabilities) -> np.ndarray:
    # sampled_probabilities: batch_size (n) x num_samples (k) x num_labels (c)
    predictive_entropy = entropy(np.mean(sampled_probabilities, axis=1))
    expected_entropy = np.mean(entropy(sampled_probabilities), axis=1)

    return predictive_entropy - expected_entropy


def calculate_bald_score_ner(sampled_probabilities) -> np.ndarray:
    # sampled_probabilities: list of num_samples (k) x batch_size (n) x num_tokens (l) x num_labels (c)
    bald_scores = []
    for i_inst in range(len(sampled_probabilities[0])):
        inst_runs = np.r_[[run[i_inst] for run in sampled_probabilities]]
        score = (
            entropy(np.mean(inst_runs, 0)).mean()
            - np.mean(entropy(inst_runs), 0).mean()
        )
        bald_scores.append(score)

    return np.array(bald_scores)


def var_ratio(sampled_probabilities):
    top_classes = np.argmax(sampled_probabilities, axis=-1)
    # count how many time repeats the strongest class
    mode_count = lambda preds: np.max(np.bincount(preds))
    modes = [mode_count(point) for point in top_classes]
    ue = 1.0 - np.array(modes) / sampled_probabilities.shape[1]
    return ue


def sampled_max_prob(sampled_probabilities):
    mean_probabilities = np.mean(sampled_probabilities, axis=1)
    top_probabilities = np.max(mean_probabilities, axis=-1)
    return 1 - top_probabilities


def probability_variance(sampled_probabilities, mean_probabilities=None):
    if mean_probabilities is None:
        mean_probabilities = np.mean(sampled_probabilities, axis=1)

    mean_probabilities = np.expand_dims(mean_probabilities, axis=1)

    return ((sampled_probabilities - mean_probabilities) ** 2).mean(1).sum(-1)


def take_idx(X_pool, idx):
    return X_pool[idx] if isinstance(X_pool, np.ndarray) else X_pool.select(idx)


def concatenate_data(
    data_1: Union[ArrowDataset, TransformersDataset],
    data_2: Union[ArrowDataset, TransformersDataset],
):
    if isinstance(data_1, ArrowDataset):
        data_1 = concatenate_datasets([data_1, data_2], info=data_1.info)
    elif isinstance(data_1, TransformersDataset):
        data_1.add(data_2)
    else:
        raise NotImplementedError
    return data_1


def _get_embeddings_wrapper(
    model_wrapper, data, model=None, data_config=None, **get_embeddings_kwargs
):
    data_config = data_config if data_config is not None else model_wrapper.data_config
    model = model_wrapper.model if model is None else model
    if model._get_name().endswith("Model"):
        prepare_model = False
    else:
        prepare_model = True
    tokenizer = model_wrapper.tokenizer
    tokenization_kwargs = dict(
        task=model_wrapper.task,
        text_name=data_config["text_name"],
        label_name=data_config["label_name"],
    )
    get_embeddings_kwargs.update(tokenization_kwargs)
    torch.manual_seed(model_wrapper.seed)

    return get_embeddings(
        model,
        data,
        tokenizer=tokenizer,
        prepare_model=prepare_model,
        **get_embeddings_kwargs,
    )


def get_similarities(
    model_name,
    X_u,
    X_l,
    sims_func="scalar_product",
    average=False,
    text_name="text",
    device="cuda",
    cache_dir=None,
    return_embeddings=False,
    batch_size: int = 100,
) -> Tuple[torch.Tensor, List[int]]:

    if cache_dir is not None:
        model_cache_dir = Path(cache_dir) / "model"
        tokenizer_cache_dir = Path(cache_dir) / "tokenizer"
    else:
        model_cache_dir = None
        tokenizer_cache_dir = None

    if X_l is not None:
        all_data = concatenate_data(X_u, X_l)
    else:
        all_data = X_u
        X_l = []

    # init models and tokenizer
    model = AutoModel.from_pretrained(model_name, cache_dir=model_cache_dir).to(device)
    tokenizer = AutoTokenizer.from_pretrained(model_name, cache_dir=tokenizer_cache_dir)
    if tokenizer.model_max_length > 5000:
        tokenizer.model_max_length = 512
    all_data = all_data.map(
        lambda x: tokenizer(x[text_name], truncation=True,),
        batched=True,
        remove_columns=all_data.column_names,
        load_from_cache_file=False,
        cache_file_name=f"tmp.data",
    )
    embeddings = get_embeddings(
        model,
        all_data,
        prepare_model=False,
        data_is_tokenized=True,
        tokenizer=tokenizer,
        use_averaging=average,
        use_automodel=True,
        batch_size=batch_size,
    )

    # TODO: seems to be a bug that must be fixed. Check whether is equal to selecting unique documents.
    embs_unique, counts = torch.unique(
        embeddings[: len(X_u)], return_inverse=True, dim=0
    )
    counts = counts.cpu().numpy().tolist()
    if len(X_l) > 0:
        embs_unique = torch.cat([embs_unique, embeddings[len(X_u) :]], dim=0)

    if sims_func == "scalar_product":
        similarities = torch.mm(embs_unique, embs_unique.T)
    elif sims_func == "cosine_similarity":
        embs_unique = embs_unique / embs_unique.norm(dim=1)[:, None]
        similarities = torch.mm(embs_unique, embs_unique.T)
    elif sims_func == "euclidean_distance":
        similarities = _get_euclidean_distances(embs_unique)
    elif sims_func == "euclidean_normalized_distance":
        embs_unique = embs_unique / embs_unique.norm(dim=1)[:, None]
        similarities = _get_euclidean_distances(embs_unique)
    elif sims_func == "mahalanobis_similarity":
        similarities = _get_mahalanobis_similarities(embs_unique)
    elif sims_func == "mahalanobis_distance":
        similarities = _get_mahalanobis_distances(embs_unique)
    else:
        raise NotImplementedError

    del model, tokenizer, embeddings
    torch.cuda.empty_cache()
    gc.collect()

    if return_embeddings:
        return similarities, counts, embs_unique

    del embs_unique
    torch.cuda.empty_cache()
    return similarities, counts


def _get_mahalanobis_similarities(embeddings):

    train_labels = np.repeat(0, len(embeddings))
    centroid, _ = compute_centroids(embeddings, train_labels, 1)
    sigma_inv, _ = compute_inv_covariance(centroid, embeddings, train_labels)
    return embeddings @ sigma_inv @ embeddings.T


def _get_mahalanobis_distances(embeddings):
    num_obj = len(embeddings)

    train_labels = np.repeat(0, num_obj)
    centroid, _ = compute_centroids(embeddings, train_labels, 1)
    sigma_inv, _ = compute_inv_covariance(centroid, embeddings, train_labels)

    # the smaller the value, the closer two objects are
    maha_scores = torch.empty(
        num_obj, num_obj, dtype=torch.float32, device=embeddings.device
    )
    batch_size = 100
    num_batches = num_obj // batch_size + int(num_obj % batch_size > 0)
    for i_batch in tqdm(range(num_batches)):
        iloc = slice(i_batch * batch_size, (i_batch + 1) * batch_size)
        batch_diff = embeddings[iloc, None, :] - embeddings[None, :, :]
        maha_scores[iloc].copy_((batch_diff @ sigma_inv * batch_diff).sum(-1))
    return -maha_scores


def _get_euclidean_distances(embeddings):
    num_obj = len(embeddings)

    train_labels = np.repeat(0, num_obj)
    centroid, _ = compute_centroids(embeddings, train_labels, 1)
    sigma_inv, _ = compute_inv_covariance(centroid, embeddings, train_labels)

    # the smaller the value, the closer two objects are
    dists = torch.empty(num_obj, num_obj, dtype=torch.float32, device=embeddings.device)
    batch_size = 100
    num_batches = num_obj // batch_size + int(num_obj % batch_size > 0)
    for i_batch in tqdm(range(num_batches)):
        iloc = slice(i_batch * batch_size, (i_batch + 1) * batch_size)
        batch_dists = torch.square(
            embeddings[iloc, None, :] - embeddings[None, :, :]
        ).sum(-1)
        dists[iloc].copy_(batch_dists)
    return -dists


def _get_similarities_from_cache_or_from_scratch(
    cache_dir, model_name, X_u, X_l, text_name, label_name, obj_id_name, device
):
    # check if we can get all stuff from cache
    if cache_dir is not None:
        similarities_path = Path(cache_dir) / "embeddings"
        indexes_path = Path(cache_dir) / "indexes"
        if similarities_path.is_file() and indexes_path.is_file():
            idx2obj = pickle.load(indexes_path.open("rb"))
            obj2idx = {obj: idx for idx, obj in idx2obj.items()}
            similarities = torch.load(similarities_path)
            return similarities, idx2obj, obj2idx

    all_data = concatenate_data(X_u, X_l)
    idx2obj = {i: all_data[i][obj_id_name] for i in range(len(all_data))}
    obj2idx = {obj: idx for idx, obj in idx2obj.items()}

    # init models and tokenizer
    model = AutoModel.from_pretrained(model_name).to(device)
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    all_data = all_data.map(
        lambda x: tokenizer(x[text_name], truncation=True,),
        batched=True,
        remove_columns=all_data.column_names,
    )
    embeddings = get_embeddings(
        model,
        all_data,
        prepare_model=False,
        data_is_tokenized=True,
        tokenizer=tokenizer,
        use_automodel=True,
    )
    # get similarities from embeddings
    # https://stackoverflow.com/questions/50411191/how-to-compute-the-cosine-similarity-in-pytorch-for-all-rows-in-a-matrix-with-re
    embs_norm = embeddings / embeddings.norm(dim=1)[:, None]
    similarities = torch.mm(embs_norm, embs_norm.T)

    # save to cache if path is provided
    if cache_dir is not None:
        similarities_path = Path(cache_dir) / "embeddings"
        indexes_path = Path(cache_dir) / "indexes"
        pickle.dump(idx2obj, indexes_path.open("wb"))
        torch.save(similarities, similarities_path)

    return similarities, idx2obj, obj2idx


def calculate_unicentroid_mahalanobis_distance(embs_unique, labeled_indices):
    train_embeddings = embs_unique[labeled_indices]
    train_labels = np.repeat(0, len(labeled_indices))

    centroids, centroids_mask = compute_centroids(train_embeddings, train_labels, 1)
    sigma_inv, _ = compute_inv_covariance(centroids, train_embeddings, train_labels)
    dists = mahalanobis_distance_with_known_centroids_sigma_inv(
        centroids, centroids_mask, sigma_inv, embs_unique,
    )
    closest_centroid_distance = dists.min(dim=1)[0]
    return closest_centroid_distance


def calculate_mahalanobis_distance(
    model_wrapper,
    train_data: Union[ArrowDataset, TransformersDataset] = None,
    unlabeled_data: Union[ArrowDataset, TransformersDataset] = None,
    embeddings: torch.Tensor = None,  # if provided, expected to be the emb-s of train followed by emb-s of unlabeled
    classwise: bool = True,  # use an own centroid for each class
    batched: bool = False,  # whether to recalculate centroids & sigma_inv after each iteration
    outlier_ids: Union[
        np.ndarray, List[int]
    ] = None,  # necessary when `batched=True` to filter outlers
    n_instances: Union[
        int, None
    ] = None,  # necessary when `batched=True` to know how many instances to select
    use_triplet: bool = False,  # whether calculate Mahalanobis Triplet scores
    unicentroid_dists: torch.Tensor = None,  # used when `use_triplet`, otherwise ignored
    triplet_lambda: float = 0.25,  # used when `use_triplet`, otherwise ignored
    use_v2: bool = False,
    use_activation: bool = False,
    use_spectralnorm: bool = True,
    model=None,  # if provided, is used instead of `model_wrapper.model`
    data_is_tokenized: bool = False,
    data_config=None,
    batch_size: int = 100,
    to_numpy: bool = True,
) -> Union[Union[torch.Tensor, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
    func_for_inv_cov = (
        compute_inv_covariance if not use_v2 else compute_inv_covariance_v2
    )
    func_for_dists_calc = (
        mahalanobis_distance_with_known_centroids_sigma_inv
        if not use_v2
        else mahalanobis_distance_with_known_centroids_sigma_inv_v2
    )
    if use_triplet:
        assert (
            0 <= triplet_lambda <= 1
        ), f"Lambda (triplet_lambda) must be between 0 and 1 but received the value {triplet_lambda}"

    # Calculate the embeddings if not provided
    if embeddings is None:
        all_data = concatenate_data(train_data, unlabeled_data)
        embeddings = _get_embeddings_wrapper(
            model_wrapper,
            all_data,
            model,
            data_config,
            use_activation=use_activation,
            use_spectralnorm=use_spectralnorm,
            data_is_tokenized=data_is_tokenized,
            batch_size=batch_size,
            to_numpy=False,
        )

    if use_triplet and unicentroid_dists is None:
        unicentroid_dists = calculate_mahalanobis_distance(
            model_wrapper,
            train_data,
            unlabeled_data,
            embeddings=embeddings,
            classwise=False,
            batched=False,
            model=model,
            use_v2=use_v2,
            use_activation=use_activation,
            use_spectralnorm=use_spectralnorm,
            data_is_tokenized=data_is_tokenized,
            data_config=data_config,
            batch_size=batch_size,
            to_numpy=False,
        ).squeeze(1)

    train_embeddings = embeddings[: len(train_data)]
    eval_embeddings = embeddings[len(train_data) :]
    if classwise:
        label_name = (
            data_config["label_name"]
            if data_config is not None
            else model_wrapper.data_config["label_name"]
        )
        train_labels = np.array(train_data[label_name])
        num_labels = model_wrapper.num_labels
    else:
        train_labels = np.zeros(len(train_data))
        num_labels = 1

    if not batched:
        centroids, centroids_mask = compute_centroids(
            train_embeddings, train_labels, num_labels
        )
        sigma_inv, _ = func_for_inv_cov(centroids, train_embeddings, train_labels)
        dists = func_for_dists_calc(
            centroids, centroids_mask, sigma_inv, eval_embeddings,
        )
        if use_triplet:
            closest_centroid_distance = dists.min(dim=1)[0]
            dists = (
                closest_centroid_distance * triplet_lambda
                - (1 - triplet_lambda) * unicentroid_dists
            )
        if to_numpy:
            return dists.cpu().detach().numpy()
        return dists
    else:
        if not classwise:
            raise ValueError(
                "Non-classwise Mahalanobis does not need to be run batched! Check your configuration."
            )
        id_queries = []
        queries_uncertainty = []
        for i in tqdm(range(n_instances), desc="Queries done:"):
            centroids, centroids_mask = compute_centroids(
                train_embeddings, train_labels, num_labels
            )
            sigma_inv, _ = func_for_inv_cov(centroids, train_embeddings, train_labels)
            dists = func_for_dists_calc(
                centroids, centroids_mask, sigma_inv, eval_embeddings,
            )
            closest_centroid_distance = dists.min(dim=1)[
                0
            ]  # taking zero el. since torch.min returns (vals, ids)
            if use_triplet:
                scores = (
                    closest_centroid_distance * triplet_lambda
                    - (1 - triplet_lambda) * unicentroid_dists
                )
            else:
                scores = closest_centroid_distance
            scores[id_queries] = -float("inf")
            if outlier_ids is not None:
                scores[outlier_ids] = -float("inf")
            id_query = scores.argmax().item()
            id_queries.append(id_query)
            queries_uncertainty.append(scores[id_query].item())
            train_embeddings = torch.cat(
                [train_embeddings, eval_embeddings[id_query].unsqueeze(0)], dim=0
            )
            train_labels = np.append(train_labels, unlabeled_data[id_query][label_name])
        uncertainty_estimates = scores.cpu().detach().numpy()
        uncertainty_estimates[id_queries] = queries_uncertainty
        return np.array(id_queries), uncertainty_estimates


def calculate_mahalanobis_triplet_scores(
    model_wrapper,
    train_data: Union[ArrowDataset, TransformersDataset],
    unlabeled_data: Union[ArrowDataset, TransformersDataset],
    lamb: float = 0.25,
    use_finetuned: bool = True,
    batched: bool = False,  # whether to recalculate centroids & sigma_inv after each iteration
    n_instances: Union[
        int, None
    ] = None,  # necessary when `batched=True` to know how many instances to select
    to_numpy: bool = False,
    **mahalanobis_kwargs,
):
    assert (
        0 <= lamb <= 1
    ), f"Lambda (lamb) must be between 0 and 1 but received the value {lamb}"

    if not use_finetuned:
        checkpoint_name = model_wrapper.model.name_or_path
        try:
            model = AutoModel.from_pretrained(checkpoint_name).cuda()
        except HTTPError:
            model = AutoModel.from_pretrained(
                checkpoint_name, local_files_only=True
            ).cuda()

        unicentroid_dists = calculate_mahalanobis_distance(
            model_wrapper,
            train_data,
            unlabeled_data,
            classwise=False,
            batched=False,
            model=model,
            to_numpy=False,
            **mahalanobis_kwargs,
        ).squeeze(1)
        mahalanobis_kwargs.update(
            dict(
                use_triplet=True,
                triplet_lambda=lamb,
                batched=batched,
                n_instances=n_instances,
                unicentroid_dists=unicentroid_dists,
            )
        )
    else:
        mahalanobis_kwargs.update(
            dict(
                use_triplet=True,
                triplet_lambda=lamb,
                batched=batched,
                n_instances=n_instances,
            )
        )

    maha_triplet_output = calculate_mahalanobis_distance(
        model_wrapper,
        train_data,
        unlabeled_data,
        classwise=True,  # use an own centroid for each class
        to_numpy=False,
        **mahalanobis_kwargs,
    )
    if not batched and to_numpy:
        return maha_triplet_output.cpu().detach().numpy()
    return maha_triplet_output


def calculate_triplet_scores(
    model_wrapper,
    train_data: Union[ArrowDataset, TransformersDataset],
    unlabeled_data: Union[ArrowDataset, TransformersDataset],
    strategy: str = "lc",  # strategy to use for uncertainty calculation
    lamb: float = 0.25,
    scale_distances: bool = True,
    use_finetuned: bool = True,
    **mahalanobis_kwargs,
):
    assert (
        0 <= lamb <= 1
    ), f"Lambda (lamb) must be between 0 and 1 but received the value {lamb}"

    if not use_finetuned:
        checkpoint_name = model_wrapper.model.name_or_path
        try:
            model = AutoModel.from_pretrained(checkpoint_name).cuda()
        except HTTPError:
            model = AutoModel.from_pretrained(
                checkpoint_name, local_files_only=True
            ).cuda()
    else:
        model = None

    unicentroid_dists = calculate_mahalanobis_distance(
        model_wrapper,
        train_data,
        unlabeled_data,
        classwise=False,
        batched=False,
        model=model,
        to_numpy=True,
        **mahalanobis_kwargs,
    ).squeeze(1)
    if scale_distances:
        unicentroid_dists = unicentroid_dists / unicentroid_dists.max()

    if strategy == "lc":
        uncertainty_scores = 1 - model_wrapper.predict_proba(unlabeled_data).max(axis=1)
    elif strategy == "entropy":
        probas = model_wrapper.predict_proba(unlabeled_data)
        uncertainty_scores = np.sum(-probas * np.log(probas), axis=1)
    elif strategy == "margin":
        probas = model_wrapper.predict_proba(unlabeled_data)
        probas.sort(axis=1)
        max_probas = probas[:, -1]
        second_max_probas = probas[:, -2]
        uncertainty_scores = 1 + second_max_probas - max_probas
    else:
        raise NotImplementedError

    scores = uncertainty_scores * lamb - (1 - lamb) * unicentroid_dists
    return scores


def calculate_mahalanobis_filtering_scores(
    model_wrapper,
    train_data: Union[ArrowDataset, TransformersDataset],
    unlabeled_data: Union[ArrowDataset, TransformersDataset],
    strategy: str = "lc",  # strategy to use for uncertainty calculation
    filtering_share: float = 0.01,
    use_finetuned: bool = True,
    **mahalanobis_kwargs,
):
    assert (
        0 <= filtering_share < 1
    ), f"Parameter filtering share must be between 0 and 1 but received the value {filtering_share}"

    if not use_finetuned:
        checkpoint_name = model_wrapper.model.name_or_path
        try:
            model = AutoModel.from_pretrained(checkpoint_name).cuda()
        except HTTPError:
            model = AutoModel.from_pretrained(
                checkpoint_name, local_files_only=True
            ).cuda()
    else:
        model = None

    batched = mahalanobis_kwargs.pop("batched")

    unicentroid_dists = calculate_mahalanobis_distance(
        model_wrapper,
        train_data,
        unlabeled_data,
        classwise=False,
        batched=False,
        model=model,
        to_numpy=True,
        **mahalanobis_kwargs,
    ).squeeze(1)

    num_to_filter = round(filtering_share * len(unicentroid_dists))
    outlier_ids = np.argsort(unicentroid_dists)[-num_to_filter:]

    if strategy == "lc":
        uncertainty_scores = 1 - model_wrapper.predict_proba(unlabeled_data).max(axis=1)
    elif strategy == "entropy":
        probas = model_wrapper.predict_proba(unlabeled_data)
        uncertainty_scores = np.sum(-probas * np.log(probas), axis=1)
    elif strategy == "margin":
        probas = model_wrapper.predict_proba(unlabeled_data)
        probas.sort(axis=1)
        max_probas = probas[:, -1]
        second_max_probas = probas[:, -2]
        uncertainty_scores = 1 + second_max_probas - max_probas
    elif strategy == "mahalanobis":
        maha_output = calculate_mahalanobis_distance(
            model_wrapper,
            train_data,
            unlabeled_data,
            classwise=True,  # use an own centroid for each class
            batched=batched,
            outlier_ids=outlier_ids,
            to_numpy=True,
            **mahalanobis_kwargs,
        )
        if not batched:
            uncertainty_scores = maha_output.min(axis=1)
        else:
            return maha_output
    else:
        raise NotImplementedError
    if not batched:
        uncertainty_scores[outlier_ids] = -float("inf")

    return uncertainty_scores


def calculate_ddu_scores(
    model_wrapper,
    data_train,
    data_test,
    use_activation: bool = False,
    use_spectralnorm: bool = True,
    data_is_tokenized=False,
    data_config=None,
    batch_size=100,
    to_numpy=True,
):
    start_time = time.time()

    data_config = data_config if data_config is not None else model_wrapper.data_config
    kwargs = dict(
        # General
        model=model_wrapper.model,
        prepare_model=True,
        batch_size=batch_size,
        to_numpy=False,
        # DDU
        use_activation=use_activation,
        use_spectralnorm=use_spectralnorm,
        # Tokenization
        data_is_tokenized=data_is_tokenized,
        tokenizer=model_wrapper.tokenizer,
        task=model_wrapper.task,
        text_name=data_config["text_name"],
        label_name=data_config["label_name"],
    )

    train_embeddings = get_embeddings(dataloader_or_data=data_train, **kwargs)
    test_embeddings = get_embeddings(dataloader_or_data=data_test, **kwargs)
    train_labels = _get_labels(data_train, data_config)

    # # TODO: remove this step
    # pca = PCA(n_components=32)
    # device = train_embeddings.device
    # train_embeddings = torch.Tensor(
    #     pca.fit_transform(train_embeddings.cpu().detach().numpy())
    # ).to(device)
    # test_embeddings = torch.Tensor(
    #     pca.transform(test_embeddings.cpu().detach().numpy())
    # ).to(device)
    # # End

    gmm, jitter = gmm_fit(train_embeddings, train_labels)
    label_probs = torch.Tensor(np.bincount(train_labels) / len(train_labels)).to(
        train_embeddings.device
    )
    assert torch.all(label_probs > 0), "All labels must present in the training sample!"

    log_probs = get_gmm_log_probs(gmm, test_embeddings)
    scores = compute_density(log_probs, label_probs)
    return probably_to_cpu_and_calculate_time(
        scores, to_numpy, start_time, model_wrapper.time_dict_path, model_wrapper.name
    )


def _get_labels(data, data_config):
    if isinstance(data, DataLoader):
        data = data.dataset
    return np.array(data[data_config["label_name"]])


def probably_to_cpu_and_calculate_time(
    scores, to_numpy, start_time, time_dict_path, model_name
):
    if to_numpy:
        scores = scores.cpu().detach().numpy()

    time_work = time.time() - start_time
    time_dict = json_load(time_dict_path)
    time_dict[model_name + "_predict"].append(time_work)
    json_dump(time_dict, time_dict_path)
    return scores


def calculate_ddu_scores_cv(
    model_wrapper,
    data_train,
    data_test,
    use_activation: bool = False,
    use_spectralnorm: bool = True,
    data_is_tokenized=False,
    data_config=None,
    batch_size=100,
    to_numpy=True,
):
    start_time = time.time()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    train_loader = DataLoader(
        data_train,
        shuffle=False,
        batch_size=batch_size,  # 64
        pin_memory=(str(device).startswith("cuda")),
        collate_fn=torch_default_data_collator,
        num_workers=0,
    )
    pool_loader = torch.utils.data.DataLoader(
        data_test,
        batch_size=128,
        shuffle=False,
        pin_memory=(str(device).startswith("cuda")),
        collate_fn=torch_default_data_collator,
        num_workers=0,
    )
    num_classes = 10
    model = model_wrapper.model
    model.eval()
    embeddings, labels = get_embeddings_(
        model,
        train_loader,
        num_dim=512,
        dtype=torch.double,
        device=device,
        storage_device=device,
    )
    gaussians_model, jitter_eps = gmm_fit_(
        embeddings=embeddings, labels=labels, num_classes=num_classes
    )
    print("Gmm training ended ========================================")
    print("Performing acquisition ========================================")
    model.eval()
    class_prob = class_probs(train_loader)
    logits, labels = gmm_evaluate(
        model,
        gaussians_model,
        pool_loader,
        device=device,
        num_classes=num_classes,
        storage_device=device,
    )
    scores = compute_density_(logits, class_prob)
    if to_numpy:
        scores = scores.cpu().detach().numpy()

    time_work = time.time() - start_time
    time_dict = json_load(model_wrapper.time_dict_path)
    time_dict[model_wrapper.name + "_predict"].append(time_work)
    json_dump(time_dict, model_wrapper.time_dict_path)

    return scores


def get_embeddings_(
    net,
    loader: torch.utils.data.DataLoader,
    num_dim: int,
    dtype,
    device,
    storage_device,
    batch_size=100,
):
    num_samples = len(loader.dataset)
    embeddings = torch.empty((num_samples, num_dim), dtype=dtype, device=storage_device)
    labels = torch.empty(num_samples, dtype=torch.int, device=storage_device)

    with torch.no_grad():
        start = 0
        for batch in tqdm(loader):
            data = batch["image"].to(device)
            label = batch["labels"].to(device)

            if isinstance(net, nn.DataParallel):
                net.module(data, label)
                out = net.module.feature
            else:
                net(data, label)
                out = net.feature

            end = start + len(data)
            embeddings[start:end].copy_(out, non_blocking=True)
            labels[start:end].copy_(label, non_blocking=True)
            start = end

    return embeddings, labels


def get_X_pool_subsample(X_pool, subsample: Union[int, float], seed: int):
    subsample_length = round(len(X_pool) / subsample)
    np.random.seed(seed)
    subsample_indices = np.random.choice(range(len(X_pool)), subsample_length, False)
    X_pool_subsample = X_pool.select(subsample_indices)
    return X_pool_subsample, subsample_indices


def assign_ue_scores_for_unlabeled_data(
    len_unlabeled_data: int, subsample_indices: np.ndarray, subsample_scores: np.ndarray
):
    scores = np.zeros(shape=len_unlabeled_data, dtype=subsample_scores.dtype) + float(
        "inf"
    )
    scores[subsample_indices] = subsample_scores
    return scores


def filter_by_uncertainty(
    uncertainty_estimates: Union[
        np.ndarray
    ],  # the larger the uncertainty, the more valuable the example is
    uncertainty_threshold: Union[float, None],
    uncertainty_mode: str,
    n_instances: int,
    uncertainty_scores: Union[
        np.ndarray, None
    ] = None,  # probability scores obtained via - `.generate`.
    # If none, coincides with `uncertainty_estimates`.
):
    if uncertainty_scores is None:
        uncertainty_scores = uncertainty_estimates
    if uncertainty_mode == "absolute":
        cutoff_idx = np.sum(uncertainty_scores >= uncertainty_threshold)
    elif uncertainty_mode == "relative":
        cutoff_idx = ceil(len(uncertainty_estimates) * uncertainty_threshold)
    cutoff_idx = min(cutoff_idx, len(uncertainty_estimates) - n_instances)

    argsort = np.argsort(-uncertainty_estimates)
    uncertainty_estimates[argsort[:cutoff_idx]] = np.inf
    query_idx = argsort[cutoff_idx : n_instances + cutoff_idx]
    return query_idx, uncertainty_estimates


def filter_by_metric(
    uncertainty_threshold: Union[float, None],
    texts: Union[ArrowDataset, TransformersDataset],
    generated_sequences_ids: Union[List, np.ndarray],
    tokenizer,
    metric_cache_dir: Union[Path, str],
    uncertainty_mode: str = "absolute",
    uncertainty_estimates: np.ndarray = None,
    n_instances: int = None,
    metric_name: str = "sacrebleu",
    agg: str = "precision",
    modify_uncertainties: bool = True,
):
    sequences = tokenizer.batch_decode(
        generated_sequences_ids, skip_special_tokens=True
    )
    metric_name_to_load = metric_name
    kwargs = {}
    if metric_name.startswith("rouge"):
        metric_name_to_load = "rouge"
        kwargs = {"use_stemmer": True}
    elif metric_name == "sacrebleu":
        texts = [[x] for x in texts]
        metric_name = "score"

    metric = load_metric(metric_name_to_load, cache_dir=metric_cache_dir)
    metric_scores = metric.compute(
        predictions=sequences, references=texts, use_agregator=False, **kwargs
    )[metric_name]
    if metric_name_to_load == "rouge":
        metric_scores = np.array([getattr(x, agg) for x in metric_scores])

    if uncertainty_mode == "absolute":
        cutoff_idx = np.sum(metric_scores <= uncertainty_threshold)
    elif uncertainty_mode == "relative":
        cutoff_idx = ceil(len(generated_sequences_ids) * uncertainty_threshold)
    cutoff_idx = min(cutoff_idx, len(generated_sequences_ids) - n_instances)
    metric_scores_argsort = np.argsort(metric_scores)

    if not modify_uncertainties:
        return cutoff_idx

    # Make them "confident" to avoid querying them
    uncertainty_estimates[metric_scores_argsort[:cutoff_idx]] = -np.inf
    query_idx = np.argsort(-uncertainty_estimates)[:n_instances]
    # Return large scores (do we need it?)
    uncertainty_estimates[metric_scores_argsort[:cutoff_idx]] = np.inf
    return query_idx, uncertainty_estimates


def calculate_pairwise_metric_score(
    summaries: List[List[str]],
    metric_name: str = "sacrebleu",
    cache_dir=None,
    tokenizer=None,
) -> np.ndarray:
    mc_iterations = len(summaries)
    len_sample = len(summaries[0])
    # sacrebleu is normally more robust than bleu
    metric = load_metric(metric_name, cache_dir=cache_dir)
    metric_kwargs = (
        {"smooth_method": "add-k"} if metric_name == "sacrebleu" else {"smooth": True}
    )
    key_name = "score" if metric_name != "bleu" else "bleu"
    if metric_name == "bleu":
        for i in range(len(summaries)):
            for idx in range(len(summaries[i])):
                summaries[i][idx] = tokenizer.tokenize(summaries[i][idx])
    scores = []
    for idx in tqdm(range(len_sample)):
        var_sum = 0.0
        for i in range(mc_iterations):
            for j in range(mc_iterations):
                if i == j:
                    continue

                metric_value = metric.compute(
                    predictions=[summaries[i][idx]],
                    references=[[summaries[j][idx]]],
                    **metric_kwargs,
                )[key_name]
                if metric_name != "bleu":
                    metric_value = round(metric_value / 100, 5)
                var_sum += (1 - metric_value) ** 2

        scores.append(1 / (mc_iterations * (mc_iterations - 1)) * var_sum)

    return np.array(scores)


def smoothing_function(p_n, references, hypothesis, hyp_len):
    """
    Smooth-BLEU (BLEUS) as proposed in the paper:
    Chin-Yew Lin, Franz Josef Och. ORANGE: a method for evaluating automatic
    evaluation metrics for machine translation. COLING 2004.
    """
    smoothed_p_n = []
    for i, p_i in enumerate(p_n, start=1):
        # Smoothing is not applied for unigrams
        if i > 1:
            # If hypothesis length is lower than the current order, its value equals (0 + 1) / (0 + 1) = 0
            if hyp_len < i:
                assert p_i.denominator == 1
                smoothed_p_n.append(1)
            # Otherwise apply smoothing
            else:
                smoothed_p_i = (p_i.numerator + 1) / (p_i.denominator + 1)
                smoothed_p_n.append(smoothed_p_i)
        else:
            smoothed_p_n.append(p_i)
    return smoothed_p_n


def pair_bleu(references, prediction):
    """
    Compute the bleu score between two given texts.
    A smoothing function is used to avoid zero scores when
    there are no common higher order n-grams between the
    texts.
    """
    tok_ref = [word_tokenize(s) for s in sent_tokenize(references)]
    tok_pred = [word_tokenize(s) for s in sent_tokenize(prediction)]
    score = 0
    for c_cent in tok_pred:
        try:
            score += corpus_bleu(
                [tok_ref], [c_cent], smoothing_function=smoothing_function
            )
        except KeyError:
            score = 0.0
    try:
        score /= len(tok_pred)
    except ZeroDivisionError:
        score = 0.0

    return score


def calculate_bleuvar_scores(summaries: List[List[str]]):
    """
    Given a list of generated texts, computes the pairwise BLEUvar
    between all text pairs. In addition, also finds the generation
    that has the smallest avg. BLEUvar score (most similar)
    with all other generations.
    """
    bleu_vars, min_bleuvars, min_gen_idxs, min_gens = [], [], [], []
    for sum_idx in tqdm(range(len(summaries[0]))):
        n = len(summaries)
        inst_summaries = [x[sum_idx] for x in summaries]
        bleu_scores = np.zeros((n, n), dtype=float)
        min_gen_idx = None
        min_bleuvar = float("inf")
        for j, dec_j in enumerate(inst_summaries):
            for k in range(j + 1, n):
                dec_k = inst_summaries[k]
                jk_bleu = pair_bleu(dec_j, dec_k)
                kj_bleu = pair_bleu(dec_k, dec_j)

                bleu_scores[j, k] = 1 - jk_bleu
                bleu_scores[k, j] = 1 - kj_bleu

            mu_bleuvar = np.sum(bleu_scores[j, :]) + np.sum(bleu_scores[:, j])
            if mu_bleuvar < min_bleuvar:
                min_bleuvar = mu_bleuvar
                min_gen_idx = j

        bleu_var = (bleu_scores ** 2).sum() / (n * (n - 1))
        min_gen = inst_summaries[min_gen_idx]

        bleu_vars.append(bleu_var)
        min_bleuvars.append(min_bleuvar)
        min_gen_idxs.append(min_gen_idx)
        min_gens.append(min_gen)

    return (
        np.array(bleu_vars),
        np.array(min_bleuvars),
        np.array(min_gen_idxs),
        np.array(min_gens),
    )


def choose_cm_samples(cluster_labels, query_idx, n_instances, random_query):
    # assume that we added cluster labels as coulmn in dataset
    # so we get samples, scores and clusters indices
    cluster_sizes = Counter(cluster_labels)
    new_query_idx = []
    samples_idx = []
    # split all query_idx to array by cluster label
    query_idx_by_clusters = {
        idx: list(query_idx[np.where(cluster_labels == idx)])
        for idx in list(cluster_sizes.keys())
    }

    # try faster approach - collect all indices that we will sample
    samples_per_cluster = []
    avg_sample_per_cluster = np.ceil(n_instances / len(cluster_sizes))

    sorted_cluster = [
        el[1] for el in sorted([(v, k) for k, v in cluster_sizes.items()])
    ]
    # else find smaller cluster, and sample its size, after subtract n_clusters * min_size and do same
    curr_idx = 0
    while sum(cluster_sizes.values()) > 0 and len(new_query_idx) < n_instances:
        # sample data from each cluster
        # sorted_cluster - array with cluster numbers in ascending order by size
        # cluster_labels - dict with cluster sizes as values
        curr_cluster = sorted_cluster[curr_idx]
        if cluster_sizes[curr_cluster] == 0:
            curr_idx = (curr_idx + 1) % len(sorted_cluster)
            continue
        # randomly sample from data with curr_cluster labels
        if random_query:
            sample_idx = random.choice(
                np.arange(len(query_idx_by_clusters[curr_cluster]))
            )
        else:
            sample_idx = 0
        samples_idx.append(sample_idx)
        new_query_idx.append(query_idx_by_clusters[curr_cluster][sample_idx])
        # and remove this sample from data
        query_idx_by_clusters[curr_cluster] = np.delete(
            query_idx_by_clusters[curr_cluster], sample_idx
        )
        # after subtract 1 from this cluster
        cluster_sizes[curr_cluster] -= 1
        curr_idx = (curr_idx + 1) % len(sorted_cluster)
    return np.array(new_query_idx), samples_idx


def get_num_query_instances_or_tokens(
    config: DictConfig,
    initial_data: Union[list, TransformersDataset, ArrowDataset],
    unlabeled_data: Union[list, TransformersDataset, ArrowDataset],
    framework: str = "transformers",
    tokens_column_name: str or None = None,
) -> int:
    """
    Function to get number of insatnces / tokens to query
    :param config: config.al from original config
    :param tokens_column_name: only used when `framework == "transformers"`
    """
    if isinstance(config.step_p_or_n, int):
        return config.step_p_or_n
    elif (
        "split_by_tokens" in config and config.split_by_tokens
    ):  # and config.al.step_p_or_n is of type float
        all_data = list(initial_data) + list(unlabeled_data)
        total_num_tokens = sum([len(x[tokens_column_name]) for x in all_data])
        return round(config.step_p_or_n * total_num_tokens)
    else:
        len_train = len(initial_data) + len(unlabeled_data)
        return round(config.step_p_or_n * len_train)


def calculate_actune_scores(
    model,
    data_test,
    probas,
    uncertainty_estimates,
    n_instances,
    use_activation: bool = False,
    use_spectralnorm: bool = False,
    data_is_tokenized=False,
    data_config=None,
    batch_size=100,
    k=10,
    lamb=0.5,
):
    pseudo_classes = np.argmax(probas, axis=-1)

    data_config = data_config if data_config is not None else model.data_config
    kwargs = dict(
        # General
        model=model.model,
        prepare_model=True,
        batch_size=batch_size,
        to_numpy=False,
        # DDU
        use_activation=use_activation,
        use_spectralnorm=use_spectralnorm,
        # Tokenization
        data_is_tokenized=data_is_tokenized,
        tokenizer=model.tokenizer,
        task=model.task,
        text_name=data_config["text_name"],
        label_name=data_config["label_name"],
    )

    test_embeddings = (
        get_embeddings(dataloader_or_data=data_test, **kwargs).detach().cpu()
    )

    km = KMeans(n_clusters=k).fit(test_embeddings)

    cluster_unc = []

    for clust_lbl in range(k):
        cluster_ids = np.where(km.labels_ == clust_lbl)[0]
        cluster_class = pseudo_classes[cluster_ids]
        cur_cluster_div = 0
        for i in np.unique(cluster_class):
            cur_ids = np.where(cluster_class == i)[0]
            freq = len(cur_ids) / len(cluster_class)
            cur_cluster_div += freq * np.log(freq)
        cluster_unc.append(
            np.sum(uncertainty_estimates[cluster_ids]) / len(cluster_ids)
            - lamb * cur_cluster_div
        )

    cl_argsort = np.argsort(np.array(cluster_unc))

    rez = []
    for i in cl_argsort:
        cluster_ids = np.where(km.labels_ == i)[0]
        samples_from_cluster = n_instances // k
        argsort = np.argsort(-uncertainty_estimates[cluster_ids])[:samples_from_cluster]
        rez.extend(cluster_ids[argsort])

    return np.array(rez), uncertainty_estimates


def _best_versus_second_best(proba):
    ind = np.argsort(proba)
    return proba[ind[-1]] - proba[ind[-2]]



from typing import Union
from datasets import Dataset
import numpy as np
from transformers import set_seed
import logging

from .al_strategy_utils import (
    assign_ue_scores_for_unlabeled_data,
    get_X_pool_subsample,
    calculate_bleuvar_scores,
    filter_by_uncertainty,
    calculate_pairwise_metric_score,
)
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


# https://arxiv.org/pdf/2006.08344.pdf
def bleuvar(
    model, X_pool: Union[Dataset, TransformersDataset], n_instances: int, **kwargs,
):
    mc_iterations = kwargs.get("mc_iterations", 10)
    metric_name = kwargs.get("var_metric", "bleu")
    log.info(f"Using {metric_name} metric for BLEUVar.")

    filtering_mode = kwargs.get("filtering_mode", None)
    bleuvar_threshold = kwargs.get("uncertainty_threshold", 1.0)
    uncertainty_mode = kwargs.get(
        "uncertainty_mode", "absolute"
    )  # "relative" or "absolute"
    seed = kwargs.get("seed", 42)

    if kwargs.get("subsample_size_mc_dropout", True):
        X_pool_subsample, subsample_indices = get_X_pool_subsample(
            X_pool, mc_iterations, model.seed
        )
    else:
        X_pool_subsample, subsample_indices = X_pool, np.arange(len(X_pool))

    generate_kwargs = dict(
        return_decoded_preds=True, do_sample=False, to_eval_mode=False
    )
    if kwargs.get("enable_dropout", False):
        if seed is not None:
            set_seed(seed)
        model.enable_dropout()  # model.model.train()
    else:
        model.model.eval()
        generate_kwargs["do_sample"] = True
        generate_kwargs["top_p"] = kwargs.get("generate_top_p", 0.95)

    summaries = []  # mc_iterations x len(X_pool_subsample) of str
    for _ in range(mc_iterations):
        generated_texts = model.generate(X_pool_subsample, **generate_kwargs)[
            "predictions"
        ]
        generated_texts = [
            text if len(text) > 0 else "CLS" for text in generated_texts
        ]  # fix empty texts
        summaries.append(generated_texts)

    # sacrebleu is normally more robust than bleu
    if metric_name == "bleu":
        bleu_vars = calculate_bleuvar_scores(summaries)[0]
    else:
        bleu_vars = calculate_pairwise_metric_score(
            summaries,
            metric_name=metric_name,
            cache_dir=model.cache_dir / "metrics",
            tokenizer=model.tokenizer,
        )

    if filtering_mode == "uncertainty":
        subsample_query_idx, bleu_vars = filter_by_uncertainty(
            bleu_vars, bleuvar_threshold, uncertainty_mode, n_instances
        )
    else:
        subsample_query_idx = np.argsort(-bleu_vars)[:n_instances]

    query = X_pool_subsample.select(subsample_query_idx)
    query_idx = subsample_indices[subsample_query_idx]

    uncertainty_estimates = assign_ue_scores_for_unlabeled_data(
        len(X_pool), subsample_indices, bleu_vars
    )

    return query_idx, query, uncertainty_estimates
